open Core.Std
open Geom_ops

type vertex_bouged = { origine :  vertex; position : vertex }

type poly = vertex_bouged list
type dll_poly = vertex_bouged dll

let pointe ligne p =
  match side ligne p.prev.contents.position, side ligne p.next.contents.position with
  | Gauche, Gauche
  | Droite, Droite -> true
  | _ -> false

let representant poly =
  fold_next (fun dll min ->
      well_formed dll;
      if compare_vertices dll.contents.position min.contents.position < 1 then
        dll
      else
        min)
    poly poly

let eliminate_equals l =
  Core.Std.List.dedup
    ~compare:(fun d1 d2 -> compare_vertices d1.contents.position d2.contents.position)
    l

let shortcuts ligne (lst:dll_poly list) : dll_poly list * segment list =
  let rec shortcuts ligne (lst:dll_poly list) =
    match lst with
    | [] -> [], []
    | p1 :: t ->
      if pointe ligne p1 then
        shortcuts ligne t
      else
        match t with
        | [] -> assert false
        | p2 :: t ->
          if pointe ligne p2 then
            match t with
            | [] -> assert false
            | p3 :: t ->
              let p'1 = shortcut p1 p2 in
              let p'2 = shortcut p2 p3 in
              let l1, l2 = shortcuts ligne t in
              p'1 :: p'2 :: l1,
              S (p1.contents.position,p3.contents.position) :: l2
          else
            let p' = shortcut p1 p2 in
            let l1, l2 = shortcuts ligne t in
            p' :: l1,
            S (p1.contents.position, p2.contents.position) :: l2
  in
  let polys, lignes_de_pliage = shortcuts ligne lst in
  eliminate_equals (List.map ~f:representant polys),
  lignes_de_pliage

let reposition (S(a,b)) i (S(a',b')) =
  let ab = b -= a in
  let a'b' = b' -= a' in
  let ai = i -= a in
  assert(Q.equal (scal ab ab) (scal a'b' a'b'));
  let norme_ai_sur_ab = Q.(scal ab ai / scal ab ab) in
  let a'i' = norme_ai_sur_ab *& a'b' in
  a' += a'i'

let points_d'intersection ligne (poly:dll_poly) : dll_poly list =
  let points =
    fold_next (fun node points_sur_la_ligne ->
        let point = node.contents.position in
        let point_origine = node.contents.origine in
        if sur_la_ligne ligne point then
          node :: points_sur_la_ligne
        else
          let point_suivant = node.next.contents in
          match intersect_segment (S(point,point_suivant.position)) ligne with
          | Non
          | Many -> points_sur_la_ligne
          | Vertex p when p ==/= point_suivant.position ->
            points_sur_la_ligne
          | Vertex position ->
            let a = point in
            let b = point_suivant.position in
            let a' = point_origine in
            let b' = point_suivant.origine in
            let origine = reposition (S(a,b)) position (S(a',b')) in
            insert_next node { position; origine };
            points_sur_la_ligne)
      poly []
  in
  let L(l,v) = ligne in
  let lst =
    List.sort (fun dl1 dl2 ->
        let p1 = dl1.contents.position in
        let p2 = dl2.contents.position in
        let d1 = scal v (p1 -= l) in
        let d2 = scal v (p2 -= l) in
        compare d1 d2)
      points
  in
  lst

let rajoute_les_points_sur_la_ligne ligne (poly:dll_poly) : dll_poly list * dll_poly list =
  let lst = points_d'intersection ligne poly in
  let gauche =
    List.filter ~f:(fun p ->
        match side ligne p.next.contents.position, side ligne p.prev.contents.position with
        | Gauche, _ | _, Gauche-> true
        | _ -> false)
      lst
  in
  let droite =
    List.filter ~f:(fun p ->
        match side ligne p.next.contents.position, side ligne p.prev.contents.position with
        | Droite, _ | _, Droite-> true
        | _ -> false)
      lst
  in
  gauche, droite

let gauchounet ligne poly =
  List.exists ~f:(fun v -> side ligne v.position = Gauche) poly

let cut_out (point:dll_poly) =
  point.prev.next <- point.next;
  point.next.prev <- point.prev

let normalize_poly poly =
  let l =
    fold_next (fun point l ->
        let prev = point.prev.contents.position in
        let p = point.contents.position in
        let next = point.next.contents.position in
        if colineaire (prev -= p) (next -= p) then
          l
        else
          point.contents :: l)
      poly []
  in
  List.rev l

let coupe_poly ligne (poly:poly) : poly list * poly list * segment list =
  let gauche, _droite = rajoute_les_points_sur_la_ligne ligne (to_dll poly) in
  let _gauche, droite = rajoute_les_points_sur_la_ligne ligne (to_dll poly) in
  let not_pointes_gauche, not_pointes_droite =
    List.filter ~f:(fun v -> not (pointe ligne v)) gauche,
    List.filter ~f:(fun v -> not (pointe ligne v)) droite
  in
  match not_pointes_gauche, not_pointes_droite with
  | [], [] ->
    if gauchounet ligne poly then
      [poly], [], []
    else
      [], [poly], []
  | _ ->
    let polys_droite, lignes_droite = shortcuts ligne (List.rev droite) in
    let polys_gauche, lignes_gauche = shortcuts ligne gauche in
    List.map ~f:normalize_poly polys_gauche,
    List.map ~f:normalize_poly polys_droite,
    lignes_droite @ lignes_gauche

let symetrie (L(l,v)) (input:vertex_bouged) : vertex_bouged =
  let p = input.position in
  let lp = p -= l in
  let nv = normal v in
  let coef = Q.(of_int 1 / norm2 v) in
  let position = l += coef *& (scal lp v *& v -& scal lp nv *& nv) in
  { input with position }

(* let gauchounet ligne poly = *)
(*   match side ligne poly.contents.position, *)
(*         side ligne poly.next.contents.position, *)
(*         side ligne poly.prev.contents.position with *)
(*   | Gauche, _, _ | _, Gauche, _ | _, _, Gauche -> true *)
(*   | Droite, _, _ | _, Droite, _ | _, _, Droite -> false *)
(*   | _ -> assert false *)

let plie_poly ligne (poly:vertex_bouged list) =
  let gauche, droite, segs =
    coupe_poly ligne poly
  in
  let gauche, droite = Caml.List.partition (gauchounet ligne) (gauche @ droite) in
  let gauche =
    List.map ~f:(List.rev_map ~f:(symetrie ligne)) gauche
  in
  gauche, droite, segs

let plie_polys ligne polys =
  let l, s =
    List.fold_left ~f:(fun (l, s) poly ->
        let g, d, sr = plie_poly ligne poly in
        g :: d :: l, sr :: s)
      ~init:([],[]) polys
  in
  Caml.List.flatten l, Caml.List.flatten s

let plie_segments ligne segs =
  Caml.List.flatten (List.map ~f:(fun s -> plie_segment s ligne) segs)

let plie_probleme ligne (polys, skel) =
  let p, s = plie_polys ligne polys in
  p, s @ plie_segments ligne skel

let position { position } = position

let ccw (verts:vertex_bouged list) =
  Geom_ops.ccw (List.map ~f:position verts)
