open Core.Std

module Q = struct
  include Q
  include Quompare
  let sexp_of_t q = String.sexp_of_t (Q.to_string q)
  let t_of_sexp q = Q.of_string (String.t_of_sexp q)
end

let float_of_q v =
  (Z.to_float v.Q.num) /. (Z.to_float v.Q.den)

open Quompare


type vertex = { x : Q.t ; y : Q.t }
[@@ deriving sexp]
type vecteur = V of Q.t * Q.t
type ligne' = { ax : Q.t; by : Q.t; c : Q.t }
type ligne = L of vertex * vecteur
type segment = S of vertex * vertex [@@ deriving sexp]
type polygon = vertex list [@@ deriving sexp]

type mapping = (vertex (* src *) * vertex (* dst *)) list
[@@ deriving sexp]

let mapping_of_int l =
  List.map ~f:(fun ((x,y),(z,t)) ->
      ({ x = Q.of_int x ; y = Q.of_int y },
       { x = Q.of_int z ; y = Q.of_int t }))
    l

let (==/=) { x = x1 ; y = y1 } { x = x2 ; y = y2 } =
  Q.(equal x1 x2 && equal y1 y2)


let compare_vertices ({ x = x ; y = y }) ({ x = x' ; y = y' }) =
  let c = Q.compare x x' in
  if c <> 0 then c else Q.compare y y'

let compare_vertex = compare_vertices
let min_vertex a b =
  let c = compare_vertex a b in
  if Int.(c <= 0) then a else b
let max_vertex a b =
  let c = compare_vertex a b in
  if Int.(c < 0) then b else a

let compare_segment (S(p1,p2)) (S(p3,p4)) =
  let c = compare_vertex p1 p3 in
  if c <> 0 then c else compare_vertex p2 p4

module Vertex = struct
  module M = struct
    type t = vertex
      [@@deriving sexp]
      let compare = compare_vertices
      let hash = Hashtbl.hash
  end
  include M
  module Map = Map.Make(M)
  module Table = Hashtbl.Make(M)
end

let pp_vertex ppf {x; y} =
  Format.fprintf ppf "{ %a; %a }" Q.pp_print x Q.pp_print y

let pp_segment ppf (S(a, b)) =
  Format.fprintf ppf "(%a => %a)" pp_vertex a pp_vertex b

let pp_vecteur ppf (V(x, y)) =
  Format.fprintf ppf "| %a, %a |" Q.pp_print x Q.pp_print y

let pp_ligne ppf (L(p, v)) =
  Format.fprintf ppf "(%a => %a)" pp_vertex p pp_vecteur v

let orig = {x=Q.zero;y=Q.zero}

let v x y = V (Q.of_int x, Q.of_int y)
let p x y = { x = Q.of_int x ; y = Q.of_int y }

let vf x y = V (Q.of_float x, Q.of_float y)
let pf x y = { x = Q.of_float x ; y = Q.of_float y }

let (-=) ({ x = vx1 ; y = vy1 }) ({ x = vx2 ; y = vy2 }) =
  V (Q.(vx1 - vx2), Q.(vy1 - vy2))

let (+=) ({ x = vx1 ; y = vy1 }) (V (vx2, vy2)) =
  { x = Q.(vx1 + vx2) ; y = Q.(vy1 + vy2) }

let (++=) (S(s1,s2)) v =
  S(s1 += v, s2 += v)

let (+&) (V (vx1, vy1)) (V (vx2, vy2)) =
  V (Q.(vx1 + vx2), Q.(vy1 + vy2))

let (-&) (V (vx1, vy1)) (V (vx2, vy2)) =
  V (Q.(vx1 - vx2), Q.(vy1 - vy2))

let ( *&) c (V (vx, vy)) =
  V (Q.(vx * c), Q.(vy * c))

let bari l =
  orig +=
  (List.fold l ~init:(orig-=orig) ~f:(fun acc (q,v) -> acc +& (q *& (v-=orig))))

let opp_segment (S(a,b)) = (S(b,a))

let colineaire (V (vx1, vy1)) (V (vx2, vy2)) =
  let open Q in
  let c1 = (vx1 * vy2) in
  let c2 = (vx2 * vy1) in
  equal c1 c2

let vect_prod (V (v1x, v1y)) (V (v2x, v2y)) =
  Q.((v1x * v2y) - (v1y * v2x))

let norm2 (V(x, y)) = Q.(x * x + y * y)
let normal (V(x, y)) = V( Q.(~- y), x)
let scal (V(x1, y1)) (V(x2, y2)) = Q.(x1 * x2 + y1 * y2)

let is_extremum (S(v1,v2)) p =
  compare_vertices v1 p = 0 ||
  compare_vertices v2 p = 0

type 'a intersect =
  | Non
  | Many
  | Vertex of 'a
[@@ deriving sexp]

let intersect (L(p1,v1)) (L(p2,v2)) =
  let open Q in
  let { x = p1x ; y = p1y } = p1 in
  let V (v1x, v1y) = v1 in
  let { x = p2x ; y = p2y } = p2 in
  let V (v2x, v2y) = v2 in
  let col = (v1x * v2y) - (v1y * v2x) in
  if equal col (Q.of_int 0) then
    let vp = p1 -= p2 in
    if colineaire vp v1 then
      Many
    else
      Non
  else
    let t = (v2x * (p1y - p2y) + v2y * (p2x - p1x)) / col in
    Vertex (p1 += t *& v1, t)

let intersect_segment (S (p1, p2)) ligne =
  let v1 = p2 -= p1 in
  match intersect (L (p1, v1)) ligne with
  | Non -> Non | Many -> Many
  | Vertex (p, t) ->
    if Q.(geq t Q.zero && leq t Q.one) then
      Vertex p
    else
      Non

let intersect_segments (S (p1, p2)) (S (q1, q2)) =
  let v1 = p2 -= p1 in
  let w1 = q2 -= q1 in
  match intersect (L (p1, v1)) (L (q1, w1))  with
  | Non -> Non
  | Many ->
    let t p =
      Q.div (scal w1 (p -= q1)) (scal w1 w1)
    in
    let t1u = t p1 in
    let t2u = t p2  in
    let mi = Q.min t1u t2u in
    let ma = Q.max t1u t2u in
    if Q.(mi = one) then Vertex q2
    else if Q.(ma = zero) then Vertex q1
    else if Q.(zero <= ma && mi <= one) then Many
    else Non
  | Vertex (p, t) ->
    if Q.(geq t Q.zero && leq t Q.one) then
      let t' = Q.div (scal w1 (p -= q1)) (scal w1 w1) in
      if Q.(Q.zero <= t' && t' <= Q.one) then
        Vertex p
      else Non
    else
      Non

(* stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect?noredirect=1&lq=1 *)
let intersect_segments2 (S(p1,p2)) (S(q1,q2)) =
  let r = p2 -= p1 in
  let s = q2 -= q1 in
  let p = p1 in
  let q = q1 in
  let pc c = p += c *& r in
  let coef = (vect_prod r s) in
  if Q.equal coef Q.zero then begin
    let u = Q.((vect_prod (q -= p) r)) in
    if not (Q.equal u Q.zero) then Non
    else
      let t0 = Q.((scal (q -= p) r) / (scal r r)) in
      let t1 = Q.(t0 + (scal s r) / (scal r r)) in
      let t0' = Q.min t0 t1 in
      let t1' = Q.max t0 t1 in
      if Q.equal t1' Q.one
      then Vertex ( pc t1' )
      else if Q.equal t0' Q.one
      then Vertex ( pc t0' )
      else Many (* ( pc t0', pc t1') *)
  end
  else 
    let t = Q.((vect_prod (q -= p) s) / coef) in
    let u = Q.((vect_prod (q -= p) r) / coef) in
    if Q.(zero <= t  && t <= Q.one && zero <= u  && u <= one)
    then Vertex ( pc t )
    else Non

(* let intersect_segments a b = *)
(*   let r1 = intersect_segments a b in *)
(*   let r2 = intersect_segments2 a b in *)
(*   if r1 <> r2 then *)
(*     Format.printf "error for %a %a:@. %a %a@." *)
(*       pp_segment a pp_segment b *)
(*       Sexp.pp_hum ([%sexp_of: vertex intersect] r1) *)
(*       Sexp.pp_hum ([%sexp_of: vertex intersect] r2); *)
(*   r1 *)

let symetrie (L(l,v)) p =
  let lp = p -= l in
  let nv = normal v in
  let coef = Q.(of_int 1 / norm2 v) in
  l += coef *& (scal lp v *& v -& scal lp nv *& nv)

type side =
  | Gauche
  | Droite
  | Ligne

let sur_la_ligne (L(l,v)) vertex =
  colineaire v (vertex -= l)

let side line p =
  let L(l,v) = line in
  if sur_la_ligne line p then
    Ligne
  else if Q.(geq (scal (normal v) (p -= l)) (of_int 0)) then
    Gauche
  else
    Droite

let plie_segment (S(p1,p2) as segment) (L(l,v) as pli) =
  match intersect_segment segment pli with
  | Non ->
    begin match side pli p1 with
      | Gauche ->
        [S(symetrie pli p1, symetrie pli p2)]
      | Droite ->
        [S(p1, p2)]
      | Ligne ->
        assert false
    end
  | Many ->
    [segment]
  | Vertex p ->
    if not (Q.(geq (scal (normal v) (p1 -= l)) (of_int 0))) then
      [S(p1,p); S(symetrie pli p2,p)]
    else
      [S(p2,p); S(symetrie pli p1,p)]

let area vertices =
  let rec sum acc l' = match vertices, l' with
    | [], _ -> assert false
    | { x = x2 ; y = y2 } :: _, ({ x = x1 ; y = y1 } :: ([] as rest))
    | _, { x = x1 ; y = y1 } :: ({ x = x2 ; y = y2 } :: _ as rest) ->
      sum Q.(acc + (x2 - x1) * (y2 + y1)) rest
    | _, [] -> acc in
  Q.(of_int 0 - sum Q.(of_int 0) vertices / of_int 2)

let ccw vertices =
  Q.(gt (area vertices) (of_int 0))

let hull vertices =
  let { x = lx ; y = ly } as leftmost =
    List.fold_left
      ~f:(fun acc p -> if Int.(compare_vertices acc p < 0) then acc else p)
      ~init:(List.hd_exn vertices) vertices in
  let rec hull acc ({ x = cx ; y = cy } as current) prev =
    if current ==/= leftmost && acc <> [] then
      List.rev (current :: acc)
    else
      let rec find_steepest ({ x = px ; y = py } as acc) = function
        | [] -> acc
        | ({ x = x ; y = y } as p) :: rest ->
          let vp = Q.((px - cx) * (y - cy) - (py - cy) * (x - cx)) in
          let cp = Q.((px - cx) * (x - cx) + (py - cy) * (y - cy)) in
          let n = Q.((x - cx) * (x - cx) + (y - cy) * (y - cy)) in
          let np = Q.((px - cx) * (px - cx) + (py - cy) * (py - cy)) in
          if Q.(equal vp (of_int 0)) then
            if Q.(lt cp (of_int 0)) || Q.(geq n np) then
              find_steepest p rest
            else
              find_steepest acc rest
          else if Q.(gt vp (of_int 0)) then
            find_steepest p rest
          else
            find_steepest acc rest in
      let next = find_steepest prev vertices in
      hull (current :: acc) next current in
  hull [] leftmost Q.({ x = lx ; y = ly - of_int 100 })

let shape_rectangle = function
  | [[ a ; b ; c ; d ]] ->
    let v1 = b -= a in
    let v2 = c -= b in
    let v3 = d -= c in
    let v4 = a -= d in
    (* let almost_one n = *)
    (*   Q.(lt (of_int 1 - of_int 1 / of_int 1000) n && *)
    (*      lt n (of_int 1 + of_int 1 / of_int 1000)) in *)
    (* let orthog va vb = *)
    (*   let sp = Q.(va.x * vb.x + va.y * vb.y) in *)
    (*   Q.(lt (of_int 0 - of_int 1 / of_int 1000) sp && *)
    (*      lt sp (of_int 1 / of_int 1000)) in *)
    if Q.(equal zero (scal v1 v2)) &&
       Q.(equal zero (scal v2 v3)) &&
       Q.(equal zero (scal v3 v4)) then
      Some (norm2 v1,norm2 v2)
    else None
  | _ -> None

type 'a dll =
  { mutable prev : 'a dll;
    contents : 'a;
    mutable next : 'a dll; }

let singleton contents =
  let rec dll = {
    prev = dll;
    contents;
    next = dll;
  } in
  dll

let copy p =
  { p with contents = p.contents }

let well_formed dll =
  assert(phys_equal dll.next.prev dll);
  assert(phys_equal dll.prev.next dll)

let insert_next dll v =
  let node = singleton v in
  let original_next = dll.next in
  dll.next <- node;
  node.prev <- dll;
  node.next <- original_next;
  original_next.prev <- node

let fold_edges ~init ~f = function
  | [] -> init
  | [_] -> invalid_argf "fold_on_line: not a valid polygon" ()
  | first::l ->
    let rec aux acc prev = function
      | [] -> f acc prev first
      | next::l ->
        let acc = f acc prev next in
        aux acc next l
    in
    aux init first l

let forall_edges ~f = function
  | [] -> true
  | [_] -> invalid_argf "forall_edges: not a valid polygon" ()
  | first::l ->
    let rec aux prev = function
      | [] -> f prev first
      | next::l -> f prev next && aux next l
    in
    aux first l

let find_edge ~f = function
  | [] | [_] -> invalid_argf "forall_edges: not a valid polygon" ()
  | first::l ->
    let rec aux prev = function
      | [] ->
        if f prev first then
          Some (prev, first)
        else
          None
      | next::l ->
        if f prev next then
          Some (prev, next)
        else
          aux next l
    in
    aux first l

let fold_next (f:'a dll -> 'b -> 'b) (dll:'a dll) acc =
  let rec loop p acc =
    let acc = f p acc in
    if phys_equal p.next dll then
      acc
    else
      loop p.next acc
  in
  loop dll acc

let fold_prev (f:'a dll -> 'b -> 'b) (dll:'a dll) acc =
  let rec loop p acc =
    let acc = f p acc in
    if phys_equal p.prev dll then
      acc
    else
      loop p.prev acc
  in
  loop dll acc

let to_list dll =
  fold_prev (fun p acc -> p.contents :: acc) dll []

let to_dll l =
  match l with
  | [] -> failwith "to_dll liste vide"
  | h :: t ->
    let dll = ref (singleton h) in
    List.iter ~f:(fun v ->
        insert_next !dll v;
        dll := !dll.next) t;
    !dll.next

let dup_poly dll =
  to_dll (to_list dll)

let points_d'intersection ligne poly =
  let points =
    fold_next (fun node points_sur_la_ligne ->
        let point = node.contents in
        if sur_la_ligne ligne point then
          node :: points_sur_la_ligne
        else
          let point_suivant = node.next.contents in
          match intersect_segment (S(point,point_suivant)) ligne with
          | Non
          | Many -> points_sur_la_ligne
          | Vertex p when p ==/= point_suivant ->
            points_sur_la_ligne
          | Vertex p ->
            insert_next node p;
            points_sur_la_ligne)
      poly []
  in
  let L(l,v) = ligne in
  let lst =
    List.sort (fun dl1 dl2 ->
        let p1 = dl1.contents in
        let p2 = dl2.contents in
        let d1 = scal v (p1 -= l) in
        let d2 = scal v (p2 -= l) in
        compare d1 d2)
      points
  in
  lst

let rajoute_les_points_sur_la_ligne ligne poly =
  let lst = points_d'intersection ligne poly in
  let gauche =
    List.filter ~f:(fun p ->
        match side ligne p.next.contents, side ligne p.prev.contents with
        | Gauche, _ | _, Gauche-> true
        | _ -> false)
      lst
  in
  let droite =
    List.filter ~f:(fun p ->
        match side ligne p.next.contents, side ligne p.prev.contents with
        | Droite, _ | _, Droite-> true
        | _ -> false)
      lst
  in
  gauche, droite

let shortcut p1 p2 =
  (* let p1' = p1 in *)
  (* let p2' = p2 in *)
  let p1 = copy p1 in
  let p2 = copy p2 in
  p1.prev.next <- p1;
  p1.next <- p2;
  p2.prev <- p1;
  p2.next.prev <- p2;
  (* well_formed p1; *)
  (* well_formed p2; *)
  (* well_formed p1.prev; *)
  (* well_formed p2.next; *)
  (* p1'.prev <- p2'; *)
  (* p2'.next <- p1'; *)
  p1

let pointe ligne p =
  match side ligne p.prev.contents, side ligne p.next.contents with
  | Gauche, Gauche
  | Droite, Droite -> true
  | _ -> false

let representant_segment (S(p1,p2)) =
  if Int.(compare_vertex p1 p2 < 0) then
    S(p1,p2)
  else
    S(p2,p1)

let equal_segment s1 s2 =
  representant_segment s1 = representant_segment s2

let representant poly =
  fold_next (fun dll minus ->
      well_formed dll;
      if Int.(compare_vertices dll.contents minus.contents < 1) then
        dll
      else
        minus)
    poly poly

let eliminate_equals l =
  Core.Std.List.dedup
    ~compare:(fun d1 d2 -> compare_vertices d1.contents d2.contents)
    l

let plie_segments ligne segs =
  Caml.List.flatten (List.map ~f:(fun s -> plie_segment s ligne) segs)


(** union/intersect *)

module Points = struct
  open Q
  open Quompare
  type point = vertex

  let det = function
    | [| [| a00; a01 |];
	 [| a10; a11 |] |] -> 
	a00 * a11 - a01 * a10
    | [| [| a00; a01; a02 |];
	 [| a10; a11; a12 |];
	 [| a20; a21; a22 |] |] -> 
	a00*a11*a22 - a00*a12*a21 - a10*a01*a22 + 
	a10*a02*a21 + a20*a01*a12 - a20*a02*a11
    | [| [| a00; a01; a02; a03 |];
	 [| a10; a11; a12; a13 |];
	 [| a20; a21; a22; a23 |];
	 [| a30; a31; a32; a33 |] |] -> 
	a00*a11*a22*a33 - a00*a11*a23*a32 - a00*a21*a12*a33 + 
	a00*a21*a13*a32 + a00*a31*a12*a23 - a00*a31*a13*a22 - 
	a10*a01*a22*a33 + a10*a01*a23*a32 + a10*a21*a02*a33 - 
	a10*a21*a03*a32 - a10*a31*a02*a23 + a10*a31*a03*a22 + 
	a20*a01*a12*a33 - a20*a01*a13*a32 - a20*a11*a02*a33 + 
	a20*a11*a03*a32 + a20*a31*a02*a13 - a20*a31*a03*a12 - 
	a30*a01*a12*a23 + a30*a01*a13*a22 + a30*a11*a02*a23 - 
	a30*a11*a03*a22 - a30*a21*a02*a13 + a30*a21*a03*a12
    | _ -> assert false

  let ccw u v w = 
    det [| [| u.x; u.y; Q.one |];
	   [| v.x; v.y; Q.one |];
	   [| w.x; w.y; Q.one |] |] > Q.zero

  (*i DEBUG
  let ccw (xu,yu) (xv,yv) (xw,yw) = 
    eprintf "ccw((%.0f,%.0f),(%.0f,%.0f),(%.0f,%.0f)) -> " 
      xu yu xv yv xw yw;
    let r = ccw (xu,yu) (xv,yv) (xw,yw) in
    eprintf "%b\n" r; flush stderr;
    r
  i*)

  let in_circle t u v w = 
    det [| [| t.x; t.y; (t.x * t.x + t.y * t.y); Q.one |];
	   [| u.x; u.y; (u.x * u.x + u.y * u.y); Q.one |];
	   [| v.x; v.y; (v.x * v.x + v.y * v.y); Q.one |];
	   [| w.x; w.y; (w.x * w.x + w.y * w.y); Q.one |]; |] > Q.zero

  (*i DEBUG
  let in_circle (xt,yt) (xu,yu) (xv,yv) (xw,yw) = 
    eprintf "in_circle((%.0f,%.0f),(%.0f,%.0f),(%.0f,%.0f),(%.0f,%.0f)) -> " 
      xt yt xu yu xv yv xw yw;
    let r = in_circle (xt,yt) (xu,yu) (xv,yv) (xw,yw) in
    eprintf "%b\n" r; flush stderr;
    r
  i*)

end

module Triangulation = Graph.Delaunay.Make(Points)

let bbox (va,vb) v2 =
  ({x= Q.min va.x v2.x; y= Q.min va.y v2.y},
   {x= Q.max vb.x v2.x; y= Q.max vb.y v2.y})

let bbox_empty v = (v,v)

let bbox_get (va,vb) =
  (va,{x=va.x;y=vb.y},vb,{x=vb.x;y=va.y})

let norm_vertex_pair (a1,a2) =
  (min_vertex a1 a2, max_vertex a1 a2)


let delaunay l =
  let h = Hashtbl.Poly.create () in
  let add_vertex v =
      Hashtbl.Poly.set h v () in
  List.iter l ~f:(fun p -> List.iter p ~f:(fun l -> List.iter l ~f:add_vertex));
  let vs = Array.of_list (Hashtbl.Poly.keys h) in
  let tria = Triangulation.triangulate vs in
  let h = Hashtbl.Poly.create () in
  let g = Vertex.Table.create () in
  Triangulation.iter_triangles (fun a1 a2 a3 ->
      let tier = Q.div Q.one (Q.of_int 3) in
      let bari = bari [tier,a1;tier,a2;tier,a3] in
      Hashtbl.Poly.add_multi h (norm_vertex_pair (a1,a2)) (a2,a3,a1,bari);
      Hashtbl.Poly.add_multi h (norm_vertex_pair (a2,a3)) (a3,a1,a2,bari);
      Hashtbl.Poly.add_multi h (norm_vertex_pair (a3,a1)) (a1,a2,a3,bari);
      Vertex.Table.add_exn g bari (a1,a2,a3);
    ) tria;
  h,g

let cross g v1 v2 =
  let nb =
    List.fold_left g ~init:0 ~f:(fun acc l -> fold_edges l ~init:acc ~f:(fun acc p1 p2 ->
        match intersect_segments (S(p1,p2)) (S(v1,v2)) with
        | Non -> acc
        | Many -> assert false
        | Vertex _ -> acc + 1
      ))
  in
  nb mod 2 = 1

let xor a b = a <> b

let compute_exterior_point l =
  let bounding_box = ref None in
  let add_vertex v =
    (match !bounding_box with
       | None ->  bounding_box := Some (bbox_empty v);
       | Some b -> bounding_box := Some (bbox b v));
  in
  List.iter l ~f:(fun p -> List.iter p ~f:(fun l -> List.iter l ~f:add_vertex));
  (match !bounding_box with
   | None -> assert false
   | Some bbox ->
     let e = Q.of_float 0.25 in
     let (v1,_,_,_) = bbox_get bbox in
     (v1 += V(Q.neg e,Q.neg e)))


let get_shape_sides g =
  let _,baris = delaunay [g] in
  let exterior = compute_exterior_point [g] in
  let inside bari =
    List.fold g ~init:false ~f:(fun acc poly ->
        if cross [poly] bari exterior
        then ccw poly else acc)
  in
  Vertex.Table.mapi baris ~f:(fun ~key:bari ~data:(v1,v2,v3) ->
      (v1,v2,v3,inside bari)
    )


let get_shap_sides_two g1 g2 =
  let _,baris = delaunay [g1;g2] in
  let exterior = compute_exterior_point [g1;g2] in
  let inside bari g =
    List.fold g ~init:false ~f:(fun acc poly ->
        if cross [poly] bari exterior
        then ccw poly else acc)
  in
  Vertex.Table.mapi baris ~f:(fun ~key:bari ~data:(v1,v2,v3) ->
      (v1,v2,v3,(inside bari g1, inside bari g2))
    )





(*
(** compute triangulation of the set of shapes, with additional
    bounding for infinity *)
let delaunay l =
  let h = Hashtbl.Poly.create () in
  let bounding_box = ref None in
  let add_vertex v =
      (match !bounding_box with
       | None ->  bounding_box := Some (bbox_empty v);
       | Some b -> bounding_box := Some (bbox b v));
      Hashtbl.Poly.set h v () in
  List.iter l ~f:(fun p -> List.iter p ~f:(fun l -> List.iter l ~f:add_vertex));
  (match !bounding_box with
   | None -> ()
   | Some bbox ->
     let e = Q.of_float 0.25 in
     let (v1,v2,v3,v4) = bbox_get bbox in
     Hashtbl.Poly.set h (v1 += V(Q.neg e,Q.neg e)) ();
     Hashtbl.Poly.set h (v2 += V(Q.neg e,e)) ();
     Hashtbl.Poly.set h (v3 += V(e,e)) ();
     Hashtbl.Poly.set h (v4 += V(e,Q.neg e)) ();
  );
  let vs = Array.of_list (Hashtbl.Poly.keys h) in
  let tria = Triangulation.triangulate vs in
  let h = Hashtbl.Poly.create () in
  let g = Vertex.Table.create () in
  Triangulation.iter_triangles (fun a1 a2 a3 ->
      let tier = Q.div Q.one (Q.of_int 3) in
      let bari = bari [tier,a1;tier,a2;tier,a3] in
      Hashtbl.Poly.add_multi h (norm_vertex_pair (a1,a2)) (a2,a3,a1,bari);
      Hashtbl.Poly.add_multi h (norm_vertex_pair (a2,a3)) (a3,a1,a2,bari);
      Hashtbl.Poly.add_multi h (norm_vertex_pair (a3,a1)) (a1,a2,a3,bari);
      Vertex.Table.add_exn g bari (a1,a2,a3);
    ) tria;
  h,g

let side_of_triangles ~outside is_inside_a_limit h =
  (* Initialize with the limit outside, and the queue of todo *)
  let sides = Vertex.Table.create () in
  let todo = Queue.create () in
  let add_todo edge bari_old side =
    let o =
      match Hashtbl.Poly.find_exn h (norm_vertex_pair edge) with
      | [_] -> assert false
      | [(_,_,_,bari); o] when bari = bari_old -> o
      | [o;(_,_,_,bari)] when bari = bari_old -> o
      | _ -> assert false in
    Queue.enqueue todo (o, bari_old, side);
  in
  Hashtbl.Poly.iteri h ~f:(fun ~key:_ ~data ->
      match data with
      | [(v1,v2,v3,bari)] ->
        Vertex.Table.add_exn sides bari outside;
        add_todo (v1,v2) bari outside;
        add_todo (v2,v3) bari outside;
      | [_;_] -> ()
      | _ -> assert false);
  (* compute the inner triangle *)
  let rec aux () =
    match Queue.dequeue todo with
    | None -> ()
    | Some ((v1,v2,v3,bari),_,side) ->
      if not (Vertex.Table.mem sides bari) then begin
        let side = is_inside_a_limit side v1 v3 in
        Vertex.Table.add_exn sides bari side;
        add_todo (v1,v2) bari side;
        add_todo (v2,v3) bari side
      end;
      aux ()
  in
  aux ();
  sides

(**
   (* crossing *)
   let cross side bari' bari =
    let nb =
      List.fold_left g ~init:0 ~f:(fun acc l -> fold_edges l ~init:acc ~f:(fun acc p1 p2 ->
          match intersect_segments (S(p1,p2)) (S(bari',bari)) with
          | Non -> acc
          | Many -> assert false
          | Vertex _ -> acc + 1
        ))
    in
    Format.printf "bari':%a bari:%a nb:%i@."
      pp_vertex bari pp_vertex bari' nb;
    match side, nb mod 2 with
    | side, 0 -> side
    | `Outside, 1 -> `Inside
    | `Inside, 1 -> `Outside
    | _ -> assert false
   in
*)

let cross g v1 v2 =
  let nb =
    List.fold_left g ~init:0 ~f:(fun acc l -> fold_edges l ~init:acc ~f:(fun acc p1 p2 ->
        match intersect_segments (S(p1,p2)) (S(v1,v2)) with
        | Non -> acc
        | Many -> acc + 1
        | Vertex _ -> acc
      ))
  in
  nb mod 2 = 1

let xor a b = a <> b

let get_shape_sides g =
  let h,baris = delaunay [g] in
  let answer = Vertex.Table.create () in
  let iter poly =
    let is_inside_a_limit side v1 v2 =
      xor side (cross [poly] v1 v2)
    in
    let sides = side_of_triangles ~outside:false is_inside_a_limit h in
    let is_ccw = ccw poly in
    Vertex.Table.merge_into ~src:sides ~dst:answer
      ~f:(fun ~key:_ s -> function
          | None when not s -> Set_to s
          | Some s' when not s -> Set_to s'
          | _ -> (if is_ccw then Set_to true else Set_to false)
        )
  in
  List.iter g ~f:iter;
  Vertex.Table.mapi baris ~f:(fun ~key:bari ~data:(v1,v2,v3) ->
      (v1,v2,v3,Vertex.Table.find_exn answer bari)
    )

let get_shap_sides_two g1 g2 =
  let h,baris = delaunay [g1;g2] in
  let is_inside_a_limit (c1',c2') v1 v2 =
    let c1 = cross g1 v1 v2 in
    let c2 = cross g2 v1 v2 in
    (xor c1 c1', xor c2 c2')
  in
  let sides = side_of_triangles ~outside:(false,false) is_inside_a_limit h in
  Vertex.Table.mapi baris ~f:(fun ~key:bari ~data:(v1,v2,v3) ->
      (v1,v2,v3,Vertex.Table.find_exn sides bari)
    )
*)
