open Core.Std

let i'm_the_server_r = ref false
let i'm_the_server () = i'm_the_server_r := true

let we_are = "74"
let key = "74-b522506ca82b646082cb05aeb9702281"
let db = Option.value (Sys.getenv "DB") ~default:"../ICFPC2016-data"
let () =
  if not (Sys.file_exists_exn db)
  then failwithf "%s must exists: clone git@gitlab.ocamlpro.com:OCamlPro/ICFPC2016-data.git" db ()

let cache = db^"/cache/"
let problems = db^"/problems/"
let problems_spec = db^"/problems_spec/"
let solutions = db^"/solutions/"
let solutions_sent = db^"/solutions_sent/"
let solutions_bad = db^"/solutions_bad/"
let () = List.iter ~f:Unix.mkdir_p [cache; problems; solutions; solutions_sent; solutions_bad;
                                    problems_spec]

let last_request = ref (Unix.gettimeofday ())
let sleep () =
  let rec aux acc =
    if 0. < acc then
      aux (Unix.nanosleep acc)
  in
  let now = (Unix.gettimeofday ()) in
  let next = !last_request +. 1. in
  aux (next -. now);
  last_request := (Unix.gettimeofday ())


(* Fonction faisant une requete via l'API fournie *)
(* API requests are rate-limited. You must wait at least 1 second between successive API requests.
   Also, every API have their own rate limits, as described below in this page.
   When you exceed these limits, your requests will fail with 429 Rate limit exceeded. *)
let rec getIcfpcDatas ~throttled ?(post=[]) url =
  assert (!i'm_the_server_r);
  sleep ();
  let t = Curl.init () in
  let buff = Buffer.create 16384 in
  Curl.setopt t (Curl.CURLOPT_ENCODING Curl.CURL_ENCODING_GZIP);
  Curl.setopt t (Curl.CURLOPT_FOLLOWLOCATION true);
  Curl.setopt t (Curl.CURLOPT_HTTPHEADER ["Expect:";"X-API-Key: " ^ key]);
  Curl.setopt t (Curl.CURLOPT_URL url);
  Curl.setopt t (Curl.CURLOPT_HTTPGET true);
  if post <> [] then begin
    let post =
      List.map post
        ~f:(function
            | `String (field,content) ->
              Curl.CURLFORM_CONTENT(field,content,(Curl.DEFAULT))
            | `File (field,file) ->
              Curl.CURLFORM_FILECONTENT(field,file,(Curl.DEFAULT)))
    in
    Curl.setopt t (Curl.CURLOPT_HTTPPOST post)
  end;
  let writer_callback a d =
    Buffer.add_string a d;
    String.length d
  in
  Curl.set_writefunction t (writer_callback buff);
  Curl.perform t;
  let code = Curl.get_responsecode t in
  Curl.cleanup t;
  match code with
  | 429 ->
    Printf.printf ".%!";
    getIcfpcDatas ~throttled:(throttled+1) ~post url
  | x when 500 <= x && x < 600 ->
    Printf.printf "!%!";
    getIcfpcDatas ~throttled:(throttled+1) ~post url
  | _ ->
    if throttled > 0 then Printf.printf "\n%!";
    Buffer.contents buff

let getIcfpcDatas = getIcfpcDatas ~throttled:0

let getBlob blob =
  let file = cache^blob in
  if Sys.file_exists_exn file
  then In_channel.read_all file, (fun () -> ())
  else
    let contest_datas blob = "http://2016sv.icfpcontest.org/api/blob/" ^ blob in
    let data = getIcfpcDatas (contest_datas blob) in
    data, (fun () -> Out_channel.write_all file ~data)

let helloworld () =
  let url = "http://2016sv.icfpcontest.org/api/hello" in
  let res = getIcfpcDatas url in
  let res = Ezjsonm.from_string res in
  let enc = Json_encoding.(destruct (obj2 (req "ok" bool) (req "greeting" string))) in
  let ok,greeting = enc res in
  ok && (greeting = "Hello, world!")

let getLastSnapshot () =
  let url = "http://2016sv.icfpcontest.org/api/snapshot/list" in
  let res = getIcfpcDatas url in
  let res = Ezjsonm.from_string res in
  let enc = Json_encoding.(destruct (obj2 (req "ok" bool)
                                       (req "snapshots"
                                          (list (obj2 (req "snapshot_time" int)
                                                   (req "snapshot_hash" string)))))) in
  let ok,snapshots = enc res in
  if ok then
    (List.hd (List.sort (fun (t1,_) (t2,_) -> compare t2 t1) snapshots))
  else None

let () = Caml.Printexc.register_printer (fun exn ->
    try
      Some (Format.asprintf "%a"
              (Json_encoding.print_error
                 ~print_unknown:(fun _ _ -> raise Exit))
              exn)
    with Exit -> None)

type problem = {
  publish_time      : float;
  solution_size     : int;
  problem_id        : int;
  owner             : string;
  problem_size      : int;
  problem_spec_hash : string;
  ranking: (float * int) list;
} [@@deriving sexp]

type leaderboard = string * float

let getStatus () =
  match getLastSnapshot () with
  | None -> None
  | Some (_timestamp,snapshot) ->
    let datas, well_parsed = getBlob snapshot in
    let res = Ezjsonm.from_string datas in
    let enc =
      let open Json_encoding in
      let problems =
        let ranking = list (obj2 (req "resemblance" float) (req "solution_size" int)) in
        let publish_time = req "publish_time" float in
        let solution_size = req "solution_size" int in
        let problem_id = req "problem_id" int in
        let owner = req "owner" string in
        let problem_size = req "problem_size" int in
        let problem_spec_hash = req "problem_spec_hash" string in
        (req "problems"
           (list (obj7 (req "ranking" ranking)
                    publish_time solution_size problem_id
                    owner problem_size problem_spec_hash))) in
      let snapshot_time = req "snapshot_time" float in
      let leaderboard = req "leaderboard"
          (list (obj2 (req "username" string) (req "score" float)))
      in
      let users = req "users"
          (list (obj2 (req "username" string) (req "display_name" string))) in
      destruct (obj4 problems snapshot_time leaderboard users)
    in
    well_parsed ();
    let problems, _, leaderboard, _ = enc res in
    let problems = List.map problems
        ~f:(fun (ranking, publish_time, solution_size, problem_id, owner,
                 problem_size, problem_spec_hash) ->
             {
               ranking;
               publish_time;
               solution_size;
               problem_id;
               owner;
               problem_size;
               problem_spec_hash;
             }
           ) in
    Some (problems, leaderboard)


let filename_of_problem problem =
  (Printf.sprintf "%s/%07i.origami" problems problem.problem_id)
let filename_of_problem_spec problem =
  (Printf.sprintf "%s/%07i.origami" problems_spec problem.problem_id)

let writeProblem problem =
  let filename = filename_of_problem problem in
  if not (Sys.file_exists_exn filename) then begin
    Format.printf "@[writing %s@]@." filename;
    let data, well_parsed = getBlob problem.problem_spec_hash in
    well_parsed ();
    Out_channel.write_all filename ~data
  end;
  let filename = filename_of_problem_spec problem in
  let data = Sexp.to_string (sexp_of_problem problem) in
  Out_channel.write_all filename ~data

let has_perfect best p =
  (best <> None && p.owner = we_are) ||
  let has_perfect = Option.(
      best >>= fun best ->
      Int.Table.find best p.problem_id
      >>= fun r ->
      if r = 1. then Some () else None) in
  has_perfect = None

let listProblems ?best () =
  if Sys.file_exists_exn problems_spec then
    let basenames = Sys.ls_dir problems_spec in
    List.fold basenames ~init:[] ~f:(fun acc basename ->
        let file = Filename.concat problems_spec basename in
        let p = problem_of_sexp (Sexp.of_string (In_channel.read_all file)) in
        if has_perfect best p then p::acc else acc
      )
  else []

let filterProblems best probs =
  List.filter ~f:(has_perfect (Some best)) probs

let readProblem id =
  Icfp_parse.parse (In_channel.read_all (filename_of_problem id))

type resemblance = [`Good of float | `Unknown | `Perhaps of float] [@@deriving sexp]

let compare_resemblance (a:resemblance) (b:resemblance) =
  match a, b with
  | `Good f1, `Good f2 -> Float.compare f1 f2
  | `Unknown, `Unknown -> 0
  | `Perhaps f1, `Perhaps f2 -> Float.compare f1 f2
  | `Good _, _ -> -1
  | _, `Good _ -> -1
  | `Perhaps _, _ -> -1
  | _, `Perhaps _ -> -1

let is_worst ~best (b:resemblance) =
  match b with
  | `Good f2 -> Float.compare best f2 >= 0
  | `Unknown -> false
  | `Perhaps f2 -> Float.compare best f2 >= 0


type solution = {
    problem_id: int;
    solution: Solution.solution;
    resemblance: [`Good of float | `Unknown | `Perhaps of float];
  } [@@deriving sexp]

let sendSolution s =
  let data = Solution.solution_string s.solution in
  let url = "http://2016sv.icfpcontest.org/api/solution/submit" in
  let res = getIcfpcDatas
      ~post:[`String ("solution_spec",data);
             `String ("problem_id",Int.to_string s.problem_id)]
      url in
  let res = Ezjsonm.from_string res in
  let enc =
    Json_encoding.(destruct (obj6
                               (req "ok" bool)
                               (opt "problem_id" int)
                               (opt "resemblance" float)
                               (opt "solution_spec_hash" string)
                               (opt "solution_size" int)
                               (opt "error" string)
                            )) in
  let (ok,problem_id,resemblance,_,_,error) = enc res in
  if ok
  then `Ok(Option.value_exn problem_id,Option.value_exn resemblance)
  else `Error(Option.value ~default:"no error given" error)



let saveSolution_aux dir s =
  let data = Sexp.to_string (sexp_of_solution s) in
  let chk = Digest.to_hex (Digest.string data) in
  let basename = sprintf "%07i_%s.sol" s.problem_id chk in
  if not (Sys.file_exists_exn (solutions_sent^basename)) &&
     not (Sys.file_exists_exn (solutions_bad^basename)) then begin
    let file = dir^basename in
    Out_channel.write_all file ~data
  end

let saveSolution ?best s =
  let current_best =
    Option.value ~default:0.
      Option.(best >>= fun best ->
              (Int.Table.find best s.problem_id)) in
  if not (is_worst ~best:current_best s.resemblance) then
    saveSolution_aux solutions s

module Convert = struct
  type solution_new = solution [@@deriving sexp_of]

  let convert file1 dir =
    Format.printf "Convert %s@." file1;
    let s = solution_of_sexp (Sexp.of_string (In_channel.read_all file1)) in
    let s = { problem_id = s.problem_id;
              solution = s.solution;
              resemblance = s.resemblance;
            } in
    let data = Sexp.to_string (sexp_of_solution_new s) in
    let chk = Digest.to_hex (Digest.string data) in
    let basename = sprintf "%07i_%s.sol" s.problem_id chk in
    let file = dir^"/"^basename in
    Format.printf "Convert %s into %s@." file1 file;
    Out_channel.write_all file ~data

end

let backupSolution (s,f,new_resemblance) =
  Unix.rename ~src:(solutions^f) ~dst:(solutions_sent^f);
  match s.resemblance with
  | `Good _ -> ()
  | `Perhaps _ | `Unknown ->
    saveSolution_aux solutions_sent
      {s with resemblance = `Good new_resemblance}

let badSolution f =
  Unix.rename ~src:(solutions^f) ~dst:(solutions_bad^f)

let foldSolutions ~init ~f dir =
  if Sys.file_exists_exn dir then
    let sols = Sys.ls_dir dir in
    let read_one acc base =
      let file = Filename.concat dir base in
      let s = solution_of_sexp (Sexp.of_string (In_channel.read_all file)) in
      f acc base s
    in
    List.fold ~init ~f:read_one sols
  else
    init

let loadSolutions () = foldSolutions ~init:[] ~f:(fun acc base s -> (s,base)::acc) solutions
let foldSolutions_sent ~init ~f = foldSolutions ~init ~f solutions_sent

type best = float Int.Table.t

let ourBestSolutions () =
  let best = Int.Table.create () in
  let read_one () _base s =
    Int.Table.change best s.problem_id
      ~f:(fun o ->
          match o , s.resemblance with
          | Some s', `Good f when f <= s' -> o
          | _, `Good f -> Some f
          | _ -> o)
  in
  foldSolutions solutions ~init:() ~f:read_one;
  foldSolutions solutions_sent ~init:() ~f:read_one;
  foldSolutions solutions_bad ~init:() ~f:read_one;
  best

open Core_extended.Std

let git_test =
  Shell.test ~working_dir:db "git"

let git_run =
  Shell.run ~working_dir:db "git"

let rec update_db () =
  if git_test ["push"] then ()
  else begin
    git_run ["fetch"];
    git_run ["merge";"origin/master"];
    update_db ();
  end

let save_db message =
  git_run ["add";"--no-ignore-removal";"."];
  if git_test ["commit";"-am";sprintf "save db:%s" message] then begin
    Format.printf "Commit created %s@." (Shell.run_one_exn "git" ["-C";db;"rev-parse";"HEAD"]);
    update_db ()
  end
  else
    Format.printf "Nothing to save@."
