(** *)
open Core.Std

let is_no_fold best prob =
  let (shape, _skeleton) = Http_api.readProblem prob in
  let area = List.fold ~init:Q.zero shape
      ~f:(fun acc poly -> Q.add acc (Geom_ops.area poly)) in
  if Q.equal area Q.one then
    match shape with
    | [[a;b;c;d]] ->
      let solution : Solution.solution =
        {mapping = [{x=Q.zero;y=Q.zero},a;
                    {x=Q.zero;y=Q.one},b;
                    {x=Q.one;y=Q.one},c;
                    {x=Q.one;y=Q.zero},d];
         facets = [[0;1;2;3]]} in
      if Solution.check ~verbose:() solution then begin
        Http_api.saveSolution ~best {
          problem_id = prob.problem_id;
          solution;
          resemblance = `Good 1.;
        }
      end
      else
        Format.printf "@[Error is_no_fold in solution for prob:%i@]@."
          prob.problem_id
    | _ ->
      Format.printf "@[prob:%i doesn't verify the invariant!@]@." prob.problem_id


let numbers q =
  let rec aux p q =
    let p' = Q.div_2exp p 1 in
    let c = Q.compare q p' in
    if c = 0 then (p',Q.zero)
    else if c > 0 then
      (p,(* Q.div_2exp  *)Q.(p-q))
    else
      aux p' q
  in
  assert (Q.compare q Q.one <= 0);
  if Q.equal q Q.one then (Q.one,Q.zero)
  else aux Q.one q

let fold_number ~init ~f (bigstep,ministep) =
  let rec aux ministep middlestep interstep acc q =
    let c = Q.compare q Q.one in
    assert ( c <= 0 );
    if c = 0
    then acc
    else
    if Q.equal ministep Q.zero then
      let q0 = q in
      let q1 = Q.add middlestep q0 in
      let q2 = Q.add middlestep q1 in
      let acc = f acc q0 q1 Q.zero middlestep in
      let acc = f acc q1 q2 middlestep Q.zero in
      aux ministep middlestep interstep acc q2
    else
      let q0 = q in
      let q1 = Q.add middlestep q0 in
      let q2 = Q.add ministep q1 in
      let q3 = Q.add ministep q2 in
      let q4 = Q.add middlestep q3 in
      let acc = f acc q0 q1 Q.zero middlestep in
      let acc = f acc q1 q2 middlestep interstep in
      let acc = f acc q2 q3 interstep middlestep in
      let acc = f acc q3 q4 middlestep Q.zero in
      aux ministep middlestep interstep acc q4
  in
  let middlestep = Q.(bigstep - ministep) in
  let interstep = Q.(middlestep - ministep) in
  if Q.equal bigstep Q.one
  then
    if Q.equal ministep Q.zero
    then f init Q.zero Q.one Q.zero Q.one
    else
      let q0 = Q.zero in
      let q1 = middlestep in
      let q2 = Q.one in
      let acc = f init q0 q1 Q.zero middlestep in
      let acc = f acc  q1 q2 middlestep interstep in
      acc
  else
    aux ministep middlestep interstep init Q.zero

(* let fold_number_swap ~init ~f bs = *)
(*   let acc,_swap = *)
(*     fold_number ~init:(init,false) bs *)
(*       ~f:(fun (acc,swap) q1 q2 p -> *)
(*           if swap then (f acc q2 q1 p, false) else (f acc q1 q2 p, true) *)
(*         ) in *)
(*   acc *)

let is_a_rectangle best prob =
  let (shape, _skeleton) as _prob' = Http_api.readProblem prob in
  match Geom_ops.shape_rectangle shape with
  | None -> ()
  | Some (len1,len2) when Q.compare len1 Q.one <= 0 &&
                          Q.compare len2 Q.one <= 0 -> begin
    let q_perfect_root l =
      if Z.perfect_square l.Q.num && Z.perfect_square l.Q.den
      then Some (Q.make (Z.sqrt l.Q.num) (Z.sqrt l.Q.den))
      else None
    in
    match q_perfect_root len1, q_perfect_root len2 with
    | Some len1, Some len2 ->
      (* Format.printf "@[prob: %a;@]@." *)
      (*   Sexp.pp_hum (Icfp_parse.sexp_of_problem prob'); *)
      (* let (big1,small1) as d1 = numbers len1 in *)
      (* let (big2,small2) as d2 = numbers len2 in *)
      (* Format.printf "@[<hov 3>@[len1: %a; big1 %a; small1:%a@]@." *)
      (*   Q.pp_print len1 Q.pp_print big1 Q.pp_print small1; *)
      (* Format.printf "@[<hov 3>@[len2: %a; big2 %a; small2:%a@]@." *)
      (*   Q.pp_print len2 Q.pp_print big2 Q.pp_print small2; *)
      (* fold_number ~init:() d1 ~f:(fun () q1 q2 q3 q4 -> Format.printf "1: %a %a %a %a;@," *)
      (*                                Q.pp_print q1 Q.pp_print q2 Q.pp_print q3 Q.pp_print q4); *)
      (* Format.printf "@."; *)
      (* fold_number ~init:() d2 ~f:(fun () q1 q2 q3 q4 -> Format.printf "2: %a %a %a %a;@," *)
      (*                                Q.pp_print q1 Q.pp_print q2 Q.pp_print q3 Q.pp_print q4); *)
      (* Format.printf "@."; *)
      let to_dst = match shape with
        | [[dst1;dst2;_dst3;dst4]] ->
          let open Geom_ops in
          let u = dst2 -= dst1 in
          let v = dst4 -= dst1 in
          fun x y ->
            let x = Q.(div x len1) in
            let y = Q.(div y len2) in
            dst1 += ((x *& u)  +& (y *& v))
        | _ -> assert false in
      let sol =
        fold_number (numbers len1) ~init:Solution.empty_solution
          ~f:(fun acc srcx1 srcx2 dstx1 dstx2 ->
              fold_number (numbers len2) ~init:acc
                ~f:(fun acc srcy1 srcy2 dsty1 dsty2 ->
                    let src : Geom_ops.polygon =
                      [{x=srcx1;y=srcy1};{x=srcx2;y=srcy1};{x=srcx2;y=srcy2};{x=srcx1;y=srcy2}] in
                    let dst : Geom_ops.polygon =
                      [to_dst dstx1 dsty1;
                       to_dst dstx2 dsty1;
                       to_dst dstx2 dsty2;
                       to_dst dstx1 dsty2] in
                    Solution.add_facets (List.zip_exn src dst) acc
                  )) in
      let solution = Solution.get_solution sol in
      (* Format.printf "sol: %a@. sol:@\n%s@." *)
      (*   Sexp.pp_hum (Solution.sexp_of_solution solution) *)
      (*   (Solution.solution_string solution); *)
      if Solution.check solution then begin
        Http_api.saveSolution ~best {
          problem_id = prob.problem_id;
          solution;
          resemblance = `Good 1.;
        }
      end
      else
        Format.printf "@[Bad solution for is_a_rectangle in solution for prob:%i@]@."
          prob.problem_id

    | _ ->
      Format.printf "@[Not_perfect root id:%i len1: %a; len2: %a@]@."
        prob.problem_id Q.pp_print len1 Q.pp_print len2
    end
  | _ -> ()


(* let approx_in_unit_square best prob = *)
(*   let (shape, _skeleton) as _prob' = Http_api.readProblem prob in *)
(*   let solution : Solution.solution = *)
(*     {mapping = [{x=Q.zero;y=Q.zero},{x=Q.zero;y=Q.zero}; *)
(*                 {x=Q.zero;y=Q.one},{x=Q.zero;y=Q.one}; *)
(*                 {x=Q.one;y=Q.one},{x=Q.one;y=Q.one}; *)
(*                 {x=Q.one;y=Q.zero},{x=Q.one;y=Q.zero};]; *)
(*      facets = [[0;1;2;3]]} in *)
(*   if Solution.check ~verbose:() solution then *)
(*     let resemblance = Solution.compute_resemblance shape solution in *)
(*     Http_api.saveSolution ~best { *)
(*       problem_id = prob.problem_id; *)
(*       solution; *)
(*       resemblance; *)
(*     } *)
(*   else *)
(*     Format.printf "@[Error is_no_fold in solution for prob:%i@]@." *)
(*       prob.problem_id *)

let () =
  let best = Http_api.ourBestSolutions () in
  let problems = Http_api.listProblems ~best () in
  (** is no fold *)
  List.iter problems ~f:(is_no_fold best);
  Http_api.save_db "is_no_fold";
  let problems = Http_api.filterProblems best problems in
  (** is a rectangle *)
  List.iter problems ~f:(is_a_rectangle best);
  Http_api.save_db "is_a_rectangle"
