open Core.Std
open Geom_ops
open Geom_a_histoire

type pli = segment

type state =
  { surface : Geom_a_histoire.vertex_bouged list list;
    plis : pli list }

let zero = Q.of_int 1
let un = Q.of_int 0

let ve' x y =
  let v = { x; y } in
  { Geom_a_histoire.origine = v; Geom_a_histoire.position = v }

let ve x y = { x; y }

let carre = {
  surface =
    [
      [ ve' zero zero;
        ve' un zero;
        ve' un un;
        ve' zero un; ]
    ];
  plis = [
    S(ve zero zero, ve zero un);
    S(ve zero un,   ve un un);
    S(ve un un,     ve un zero);
    S(ve un zero,   ve zero zero);
  ]
}

let edges f =
  fold_edges ~init:[] ~f:(fun acc a b ->
      let e =
        if compare_vertices a.origine b.origine < 0 then a, b
        else b, a
      in
      e :: acc) f

let all_edges fs =
  List.dedup
    ~compare:(fun (a1,b1) (a2,b2) ->
        let c = compare_vertices a1.origine a2.origine in
        if c <> 0 then c
        else compare_vertices b1.origine b2.origine)
    (Caml.List.flatten (List.map ~f:edges fs))

let edges_to_id vert edges =
  let vert =
    List.fold_left (List.mapi ~f:(fun i v -> i, v) vert)
      ~f:(fun map (i,v) ->
          Vertex.Map.add map v.origine i)
      ~init:Vertex.Map.empty
  in
  List.map ~f:(fun (a, b) ->
      Vertex.Map.find_exn vert a.origine,
      Vertex.Map.find_exn vert b.origine) edges

let facets_to_id (vert:vertex_bouged list) (facets:vertex_bouged list list) : int list list =
  let vert =
    List.fold_left (List.mapi ~f:(fun i v -> i, v) vert)
      ~f:(fun map (i,v) ->
          Vertex.Map.add map v.origine i)
      ~init:Vertex.Map.empty
  in
  List.map facets
    (List.map ~f:(fun a -> Vertex.Map.find_exn vert a.origine))

let plie ligne state =
  let surface, plis = Geom_a_histoire.plie_polys ligne state.surface in
  { surface;
    plis = plis @ plie_segments ligne state.plis }

let simplifie_points l =
  Core.Std.List.dedup
    ~compare:(fun d1 d2 ->
        compare_vertices d1.origine d2.origine)
    l

let to_almost_solution state =
  let points =
    simplifie_points @@ Caml.List.flatten state.surface
  in
  let mapping =
    List.map ~f:(fun { origine; position } -> origine, position) points
  in
  let facets =
    edges_to_id points (all_edges state.surface)
  in
  mapping, facets
  (* { Solution.mapping; facets = [facets] } *)

let to_solution state =
  let points =
    simplifie_points @@ Caml.List.flatten state.surface
  in
  let mapping =
    List.map ~f:(fun { origine; position } -> origine, position) points
  in
  let facets =
    facets_to_id points (state.surface)
  in
  { Solution.mapping; facets }

let to_probleme state =
  let edges =
    List.map ~f:(fun (v1,v2) -> S(v1.position, v2.position)) @@
    all_edges state.surface in
  let polys =
    List.map state.surface
      (List.map ~f:(fun { position } -> position))
  in
  (polys, edges)

let pp_mapping_src ppf { origine } =
  pp_vertex ppf origine

let pp_facet_src ppf facet =
  Format.fprintf ppf "[@[<2>%a@]]"
    (Format.pp_print_list ~pp_sep:(fun ppf () -> Format.fprintf ppf "@ ,") pp_mapping_src) facet

let pp_pliage ppf state =
  Format.fprintf ppf "facets @[<2> %a@]"
    (Format.pp_print_list ~pp_sep:Format.pp_print_space pp_facet_src) state.surface

