open Geom_ops

let init () =
  Graphics.open_graph " 600x600" ;
  Graphics.auto_synchronize false

let souris = Graphics.(rgb 220 220 200)

let transform scale offset v =
  offset + int_of_float (float_of_q v *. (float scale))

let vertex_position scale (ox,oy) { x ; y } =
  transform scale ox x, transform scale oy y

let draw_polygon color scale offset l =
  let a = Array.of_list (List.map (fun v -> vertex_position scale offset v) l) in
  Graphics.set_color color;
  Graphics.fill_poly a;
  (* Graphics.set_color Graphics.black; *)
  (* Graphics.draw_poly a; *)
  ()

let show_vertex scale offset v =
  let x, y = vertex_position scale offset v in
  Graphics.fill_circle x y 3

let show_mapping_src scale offsets l =
  Graphics.set_color Graphics.black;
  List.iter (fun (src, _) -> show_vertex scale offsets src) l

let iter_round f l =
  match l with
  | [] -> assert false
  | first :: _ ->
    let rec loop l =
      match l with
      | [] -> assert false
      | [e] -> f e first
      | h1 :: ((h2 :: _) as t) ->
        f h1 h2;
        loop t
    in
    loop l


let show_segments scale offset (a,b) =
  let x, y = vertex_position scale offset a in
  Graphics.moveto x y;
  let x, y = vertex_position scale offset b in
  Graphics.lineto x y

let show_facet scale offset pos facet =
  iter_round (fun a b ->
      let a = pos.(a) in
      let b = pos.(b) in
      show_segments scale offset (a,b)
    )
    facet

let show_facets scale offset s =
  let pos = Array.of_list (List.map fst s.Solution.mapping) in
  Graphics.set_color Graphics.black;
  List.iter (show_facet scale offset pos) s.Solution.facets

let show_almost_facets scale offset mapping lst =
  let pos = Array.of_list (List.map fst mapping) in
  Graphics.set_color Graphics.black;
  List.iter (fun (a,b) ->
      let a = pos.(a) in
      let b = pos.(b) in
      show_segments scale offset (a,b)) lst

let show_polygons scale offset l =
  let ccwl, cwl =
    List.fold_left
      (fun (ccwl, cwl) v ->
         if Geom_ops.ccw v then (v :: ccwl, cwl) else (ccwl, v :: cwl))
      ([], []) l in
  List.iter (draw_polygon Graphics.yellow scale offset) ccwl;
  List.iter (draw_polygon Graphics.green scale offset) cwl

let show_skeleton scale offset s =
  Graphics.set_color Graphics.black;
  List.iter (fun (S (v1, v2)) -> show_segments scale offset (v1,v2)) s

let _show_hull scale offset l =
  let open Geom_ops in
  let all_points =
    l |>
    List.flatten |>
    List.sort_uniq compare_vertices in
  let hull =
    hull all_points |>
    List.map (vertex_position scale offset) |>
    Array.of_list in
  Graphics.set_color Graphics.green;
  Graphics.fill_poly hull

let scale_user = ref 1.

let offset_scale (offsetx, offsety) (w, h) s =
  let minx, miny, maxx, maxy =
    List.fold_left
      (fun (minx, miny, maxx, maxy) (S ({ x = x1 ; y = y1 }, { x = x2 ; y = y2 })) ->
         let minx =
           min minx (min (float_of_q x1) (float_of_q x2)) in
         let miny =
           min miny (min (float_of_q y1) (float_of_q y2)) in
         let maxx =
           max maxx (max (float_of_q x1) (float_of_q x2)) in
         let maxy =
           max maxy (max (float_of_q y1) (float_of_q y2)) in
         (minx, miny, maxx, maxy))
      (0., 0., 1., 1.) s in
  let scale =
    int_of_float @@ min
      ((1. /. (maxx -. minx)) *. float w *. !scale_user)
      ((1. /. (maxy -. miny)) *. float h *. !scale_user) in
  let offset =
    (offsetx - int_of_float (minx *. float scale)
     + (w - int_of_float ((maxx -. minx) *. float scale)) / 2,
     offsety - int_of_float (miny *. float scale)
     + (h - int_of_float ((maxy -. miny) *. float scale)) / 2) in
  offset, scale

let delaunay = ref 0
let delaunay_limit = Graphics.red
let delaunay_inner = Graphics.blue
let delaunay_point = Graphics.black

let draw_delaunay1 scale offset shape =
  let h,_ = Geom_ops.delaunay [shape] in
  Core.Std.Hashtbl.Poly.iteri h ~f:(fun ~key:(a1,b1) ~data ->
      match data with
      | [(_,_,_,bari)] ->
        if !delaunay = 2 then begin
          Graphics.set_color delaunay_point;
          show_vertex scale offset bari;
        end;
        Graphics.set_color delaunay_limit;
        show_segments scale offset (a1,b1)
      | [(_,_,_,bari1);(_,_,_,bari2)] ->
        if !delaunay = 2 then begin
          Graphics.set_color delaunay_point;
          show_segments scale offset (bari1,bari2);
        end;
        Graphics.set_color delaunay_inner;
        show_segments scale offset (a1,b1)
      | _ -> assert false
    )

let draw_delaunay2 scale offset shape =
  let sides = Geom_ops.get_shape_sides shape in
  Geom_ops.Vertex.Table.iteri sides ~f:(fun ~key:_ ~data:(v1,v2,v3,side) ->
      match side with
      | false -> ()
      | true -> draw_polygon delaunay_inner scale offset [v1;v2;v3];
    )

let skeleton = ref true

let draw_at (offsetx, offsety) (w, h) ((shape, s): Icfp_parse.problem) =
  let offset, scale = offset_scale (offsetx, offsety) (w, h) s in
  Graphics.set_color souris;
  let x1, y1 = vertex_position scale offset ({ x = Q.of_int 0 ; y = Q.of_int 0 }) in
  let x2, y2 = vertex_position scale offset ({ x = Q.of_int 1 ; y = Q.of_int 1 }) in
  Graphics.fill_rect x1 y1 (x2 - x1) (y2 - y1);
  show_polygons scale offset shape;
  if !skeleton then show_skeleton scale offset s;
  if 1 <= !delaunay &&  !delaunay <= 2 then draw_delaunay1 scale offset shape;
  if 3 <= !delaunay &&  !delaunay <= 3 then draw_delaunay2 scale offset shape;
  Graphics.synchronize ()



let draw (l, s) =
  Graphics.clear_graph ();
  let w = Graphics.size_x () in
  let h = Graphics.size_y () in
  let border = 10 in
  draw_at (border, border) (w - 2 * border, h - 2 * border) (l, s)

let draw_solution_at (offsetx, offsety) (w, h) (sol: Solution.solution) =
  let minx, miny, maxx, maxy = (0., 0., 1., 1.) in
  let scale =
    int_of_float @@ min
      ((1. /. (maxx -. minx)) *. float w)
      ((1. /. (maxy -. miny)) *. float h) in
  let offset =
    (offsetx - int_of_float (minx *. float scale)
     + (w - int_of_float ((maxx -. minx) *. float scale)) / 2,
     offsety - int_of_float (miny *. float scale)
     + (h - int_of_float ((maxy -. miny) *. float scale)) / 2) in
  Graphics.set_color souris;
  let x1, y1 = vertex_position scale offset ({ x = Q.of_int 0 ; y = Q.of_int 0 }) in
  let x2, y2 = vertex_position scale offset ({ x = Q.of_int 1 ; y = Q.of_int 1 }) in
  Graphics.fill_rect x1 y1 (x2 - x1) (y2 - y1);
  show_mapping_src scale offset sol.Solution.mapping;
  show_facets scale offset sol;
  Graphics.synchronize ()

let draw_solution s =
  Graphics.clear_graph ();
  let w = Graphics.size_x () in
  let h = Graphics.size_y () in
  let border = 10 in
  draw_solution_at (border, border) (w - 2 * border, h - 2 * border) s

let draw_almost_solution_at (offsetx, offsety) (w, h)
    ((mapping, segments) : (Geom_ops.mapping * (int * int) list)) =
  let minx, miny, maxx, maxy = (0., 0., 1., 1.) in
  let scale =
    int_of_float @@ min
      ((1. /. (maxx -. minx)) *. float w)
      ((1. /. (maxy -. miny)) *. float h) in
  let offset =
    (offsetx - int_of_float (minx *. float scale)
     + (w - int_of_float ((maxx -. minx) *. float scale)) / 2,
     offsety - int_of_float (miny *. float scale)
     + (h - int_of_float ((maxy -. miny) *. float scale)) / 2) in
  Graphics.set_color souris;
  let x1, y1 = vertex_position scale offset ({ x = Q.of_int 0 ; y = Q.of_int 0 }) in
  let x2, y2 = vertex_position scale offset ({ x = Q.of_int 1 ; y = Q.of_int 1 }) in
  Graphics.fill_rect x1 y1 (x2 - x1) (y2 - y1);
  show_mapping_src scale offset mapping;
  show_almost_facets scale offset mapping segments;
  Graphics.synchronize ()

let draw_almost_solution s =
  Graphics.clear_graph ();
  let w = Graphics.size_x () in
  let h = Graphics.size_y () in
  let border = 10 in
  draw_almost_solution_at (border, border) (w - 2 * border, h - 2 * border) s

let draw_solution_et_probleme (s:Solution.solution) (p:Icfp_parse.problem) =
  Graphics.clear_graph ();
  let w = Graphics.size_x () in
  let h = Graphics.size_y () in
  let border = 10 in
  let w, off_x, h, off_y =
    if w > h then w/2, w/2, h, 0 else w, 0, h/2, h/2
  in
  draw_solution_at (border, border) (w - 2 * border, h - 2 * border) s;
  draw_at (border + off_x, border + off_y) (w - 2 * border, h - 2 * border) p

exception Timeout

let wait ?(timeout = infinity) (l, s) title keys =
  let help_h = 16 * (2 + List.length keys) in
  let t0 = Unix.gettimeofday () in
  let draw () =
    Graphics.clear_graph ();
    let w = Graphics.size_x () in
    let h = Graphics.size_y () in
    let border = 10 in
    Graphics.set_color Graphics.black;
    Graphics.moveto border (h - border - 16);
    Graphics.draw_string title;
    Graphics.moveto border (h - border - 32);
    let area = List.fold_left (fun acc v ->
        Q.(+) (Geom_ops.area v) acc)
        Q.(of_int 0) l in
    Graphics.draw_string
      (Format.asprintf "Area: %s%s"
         (Q.to_string area)
         (if Geom_ops.shape_rectangle l <> None then "[RECTANGLE]" else ""));
    Graphics.set_window_title title;
    List.iteri
      (fun i (key, (help, _)) ->
         Graphics.moveto border (h - border - 16 * (i + 3)) ;
         Graphics.draw_string (Format.asprintf "%c: %s" key help))
      keys;
    draw_at (border, border) (w - 2 * border, h - help_h - 3 * border) (l, s) in
  let rec loop prev_w prev_h =
    if Unix.gettimeofday () -. t0 > timeout then raise Timeout ;
    let { Graphics.keypressed } =
      Graphics.(wait_next_event [ Poll ]) in
    let w = Graphics.size_x () in
    let h = Graphics.size_y () in
    if prev_h <> h || prev_w <> w then
      draw () ;
    if keypressed then begin
      let { Graphics.key } = Graphics.(wait_next_event [ Key_pressed ]) in
      try snd (List.assoc key keys) with
        Not_found -> (* Printf.printf "unknown key:%c\n%!" key; *) loop w h
    end else begin
      (try ignore (Unix.select [] [] [] 0.05) with Unix.Unix_error _ -> ());
      loop w h
    end in
  draw () ;
  loop (Graphics.size_x ()) (Graphics.size_y ())

let draw_vect color (sx,sy) (dx,dy) =
  Graphics.set_color color;
  Graphics.fill_circle sx sy 3;
  Graphics.moveto sx sy;
  Graphics.lineto dx dy

let show_sol_destination = ref false

let draw_sol_prob title keys sol (l, s) click =
  Graphics.clear_graph ();
  let w = Graphics.size_x () in
  let h = Graphics.size_y () in
  let border = 10 in
  Graphics.set_color Graphics.black;
  Graphics.moveto border (h - border - 16);
  Graphics.draw_string title;
  Graphics.moveto border (h - border - 32);
  let area = List.fold_left (fun acc v ->
      Q.(+) (Geom_ops.area v) acc)
      Q.(of_int 0) l in
  Graphics.draw_string
    (Format.asprintf "Area: %s%s"
       (Q.to_string area)
       (if Geom_ops.shape_rectangle l <> None then "[RECTANGLE]" else ""));
  Graphics.set_window_title title;
  List.iteri
    (fun i (key, (help, _)) ->
       Graphics.moveto border (h - border - 16 * (i + 3)) ;
       Graphics.draw_string (Format.asprintf "%c: %s" key help))
    keys;
  let w, off_x, h, off_y =
    if w > h then w/2, w/2, h, 0 else w, 0, h/2, h/2
  in
  draw_solution_at (border, border) (w - 2 * border, h - 2 * border) sol;
  let offset = (border + off_x, border + off_y) in
  let size = (w - 2 * border, h - 2 * border) in
  if !show_sol_destination then
    let dst_facets = Solution.get_destination_facets sol in
    draw_at offset size (dst_facets,[])
  else draw_at offset size (l, s);
  (match click with
   | None -> ()
   | Some (src, dst) ->
     draw_vect Graphics.black src dst;
     Graphics.synchronize ())

let offsets_size (_l,s) =
  let w = Graphics.size_x () in
  let h = Graphics.size_y () in
  let border = 10 in
  let w, off_x, h, off_y =
    if w > h then w/2, w/2, h, 0 else w, 0, h/2, h/2
  in
  let offset = (border + off_x, border + off_y) in
  let size = (w - 2 * border, h - 2 * border) in
  offset_scale offset size s

let wait_solution_et_probleme ?(timeout = infinity) sol prob title keys =
  let t0 = Unix.gettimeofday () in
  let rec loop prev_w prev_h =
    if Unix.gettimeofday () -. t0 > timeout then raise Timeout ;
    let { Graphics.keypressed } =
      Graphics.(wait_next_event [ Poll ]) in
    let w = Graphics.size_x () in
    let h = Graphics.size_y () in
    if prev_h <> h || prev_w <> w then
      ignore(draw_sol_prob title keys sol prob None);
    if keypressed then begin
      let { Graphics.key } = Graphics.(wait_next_event [ Key_pressed ]) in
      try snd (List.assoc key keys) with
        Not_found -> loop w h
    end else begin
      (try ignore (Unix.select [] [] [] 0.05) with Unix.Unix_error _ -> ());
      loop w h
    end in
  ignore(draw_sol_prob title keys sol prob None);
  loop (Graphics.size_x ()) (Graphics.size_y ())

let origin_coord scale (o_x, o_y) (x,y) =
  let scale = Q.of_int scale in
  let o_x = Q.of_int o_x in
  let o_y = Q.of_int o_y in
  let x = Q.of_int x in
  let y = Q.of_int y in
  { x = Q.((x - o_x) / scale);
    y = Q.((y - o_y) / scale) }

let interract sol prob do_stuff title keys =
  let r = ref 0 in
  let init_pos = ref None in
  let click (offset, scale) =
    let p = Graphics.mouse_pos () in
    let p' = origin_coord scale offset p in
    match !init_pos with
    | None ->
      init_pos := Some (p, p')
    | Some (_, prev) ->
      init_pos := None;
      do_stuff prev p'
  in
  let rec loop prev_w prev_h (prev_mouse_x, prev_mouse_y) prev_button =
    incr r;
    let { Graphics.keypressed; Graphics.button } =
      Graphics.(wait_next_event [ Poll ]) in
    let w = Graphics.size_x () in
    let h = Graphics.size_y () in
    let (mouse_x, mouse_y) = Graphics.mouse_pos () in
    if not button && prev_button then
      click (offsets_size !prob);
    let vect = match !init_pos with
      | None -> None
      | Some (p,_) -> Some (p, (mouse_x, mouse_y))
    in
    if prev_h <> h || prev_w <> w || prev_mouse_x <> mouse_x || prev_mouse_y <> mouse_y
       || button <> prev_button then
      draw_sol_prob title keys !sol !prob vect;
    if keypressed then begin
      let { Graphics.key } = Graphics.(wait_next_event [ Key_pressed ]) in
      match snd (List.assoc key keys) with
      | exception Not_found ->
        loop w h (mouse_x, mouse_y) button
      | f ->
        f ();
        draw_sol_prob title keys !sol !prob vect;
        loop w h (mouse_x, mouse_y) button
    end else begin
      (try ignore (Unix.select [] [] [] 0.05) with Unix.Unix_error _ -> ());
      loop w h (mouse_x, mouse_y) button
    end in
  draw_sol_prob title keys !sol !prob None;
  loop (Graphics.size_x ()) (Graphics.size_y ()) (Graphics.mouse_pos ()) (Graphics.button_down ())
