open Core.Std
open Geom_ops
open Quompare
(*
    the source positions part
    the facets part
    the destination positions part
*)

(*
The vertices of an origami are indexed by integers starting from 0.

The source positions part consists of the following items, each in one line:

    the number of vertices

    for each vertex, its coordinate
*)

let add_int b i = Buffer.add_string b (string_of_int i)
let add_bint b i = Buffer.add_string b (Z.to_string i)

let add_num b v =
  if Z.equal v.Q.den Z.one then
    add_bint b v.Q.num
  else
    let () = add_bint b v.Q.num in
    Buffer.add_char b '/';
    add_bint b v.Q.den

let add_vertex b p =
  add_num b p.x;
  Buffer.add_char b ',';
  add_num b p.y

let add_vertices b l =
  let n = List.length l in
  add_int b n;
  Buffer.add_char b '\n';
  List.iter ~f:(fun v ->
      add_vertex b v;
      Buffer.add_char b '\n')
    l

let add_facet b l =
  let n = List.length l in
  add_int b n;
  Buffer.add_char b ' ';
  Buffer.add_string b (String.concat ~sep:" " (List.map ~f:string_of_int l))

let add_facets b l =
  let n = List.length l in
  add_int b n;
  Buffer.add_char b '\n';
  List.iter ~f:(fun v ->
      add_facet b v;
      Buffer.add_char b '\n')
    l

let add_destination_position b l =
  List.iter ~f:(fun v ->
      add_vertex b v;
      Buffer.add_char b '\n')
    l

type solution = {mapping: Geom_ops.mapping; facets: int list list}
[@@ deriving sexp]

let solution_string s =
  let b = Buffer.create 1024 in
  add_vertices b (List.map ~f:fst s.mapping);
  add_facets b s.facets;
  add_destination_position b (List.map ~f:snd s.mapping);
  Buffer.contents b

let solution_size_string s =
  let r = ref 0 in
  String.iter ~f:(function
      | ' ' | '\n' | '\r' | '\t' -> ()
      | _ -> incr r) s;
  !r

let solution_size s =
  solution_size_string (solution_string s)

let no_duplicated l =
  List.sort ~cmp:compare l = List.dedup l

let check ?verbose (sol:solution) =
  let verbose = verbose <> None in
  let ( *** ) s b =
    if verbose && not b then Format.eprintf "@[Error invariant broken: %s@]@." s;
    b in
  let ( **** ) s v =
    match v with
    | None -> true
    | Some (src, dst) ->
      if verbose then
        Format.eprintf "@[Error invariant broken: %s by mapping %a -> %a@]@." s
          Geom_ops.pp_vertex src
          Geom_ops.pp_vertex dst;
      false
  in
  begin
    "All the source positions of the vertices are within the initial \
     square spanned by the four vertices (0,0), (1,0), (1,1), (0,1)."
    ****
    List.find sol.mapping ~f:(fun (src,_) ->
        not
          (Q.zero <= src.x && src.x <= Q.one &&
           Q.zero <= src.y && src.y <= Q.one))
    &&
    "No coordinate appears more than once in the source positions part."
    ***
    let seen = Hashtbl.Poly.create () in
    List.for_all sol.mapping ~f:(fun (src,_) ->
        match Hashtbl.Poly.add seen src () with
        | `Duplicate -> false
        | `Ok -> true)
  end
  &&
  let mapping = Int.Table.create () in
  List.iteri sol.mapping ~f:(fun i map ->
      Int.Table.add_exn mapping i map);
  begin
    "Additional, all index in facets exists"
    ***
    List.for_all sol.facets ~f:(List.for_all ~f:(
        fun id -> Int.Table.mem mapping id))
  end
  &&
  let get_src id = fst (Int.Table.find_exn mapping id) in
  let get_dst id = snd (Int.Table.find_exn mapping id) in
  begin
    "Any edge of any facet has length greater than zero. \
     Additional: the length is identical between src and dst. "
    ***
    List.for_all sol.facets ~f:(
      fun face ->
        let face = List.map face ~f:(fun id -> (id,get_src id,get_dst id)) in
        Geom_ops.forall_edges face ~f:(fun (_id1,src,dst) (_id2,src',dst') ->
            let ttsrc = src in
            let ttdst = dst in
            let src = src' -= src in
            let dst = dst' -= dst in
            let lsrc = norm2 src in
            if verbose && (Q.equal lsrc Q.zero) then
              Format.eprintf "Length zero: %a %a"
                Geom_ops.pp_vertex ttsrc
                Geom_ops.pp_vertex ttdst;
            if verbose && (not (Q.equal lsrc (norm2 dst))) then
              Format.eprintf "Length not kept: %a %a %a %a (%a<>%a)@."
                Geom_ops.pp_vertex ttsrc
                Geom_ops.pp_vertex ttdst
                Geom_ops.pp_vertex src'
                Geom_ops.pp_vertex dst'
                Q.pp_print lsrc
                Q.pp_print (norm2 dst);
            not (Q.equal lsrc Q.zero) && Q.equal lsrc (norm2 dst)
          ))
  end
  &&
  begin
    "At source positions, if two different edges share a point, the \
     point should always be one of the endpoints for both the edges. That \
     is, an edge touching another edge, or edges crossing each other are \
     prohibited."
    ***
    begin try
        let _ = List.fold sol.facets ~init:[] ~f:(
            fun acc face ->
              Geom_ops.fold_edges face ~init:acc
                ~f:(fun acc id id' ->
                    let v = Geom_ops.S((get_src id),(get_src id')) in
                    List.iter acc ~f:(fun v' ->
                        match intersect_segments v v' with
                        | Non -> ()
                        | Many when v = v' -> ()
                        | Many when v = Geom_ops.opp_segment v' -> ()
                        | Many ->
                          if verbose then
                            Format.eprintf "ovelaping segments %a %a@."
                              Geom_ops.pp_segment v
                              Geom_ops.pp_segment v';
                          raise Exit
                        | Vertex p
                          when Geom_ops.is_extremum v p &&
                               Geom_ops.is_extremum v' p
                          -> ()
                        | Vertex p ->
                          if verbose then
                            Format.eprintf "problematic segments %a %a %b %a %b@."
                              Geom_ops.pp_vertex p
                              Geom_ops.pp_segment v
                              (Geom_ops.is_extremum v p)
                              Geom_ops.pp_segment v'
                              (Geom_ops.is_extremum v' p);
                          raise Exit
                      );
                    (v::acc)
                  )
          ) in
        true;
      with Exit -> false
    end
  end
  &&
  begin
    "The size of the solution is no larger than Bs = 5000 Bytes."
    ***
    let len_string = solution_size sol in
    let b = Int.(len_string <= 5000) in
    (if verbose && not b then
       Format.eprintf "size is: %i@." len_string);
    b
  end
  &&
  begin
    "Has duplicated vertex in facet"
    ***
    List.for_all ~f:no_duplicated sol.facets
  end

type mk_solution =
  { index: (int * vertex) Vertex.Map.t;
    next_index: int;
    facets: int list list; }

let empty_solution = {index = Vertex.Map.empty; next_index = 0; facets = []}
let add_facets p s =
  let index,next_index, acc =
    List.fold ~init:(s.index,s.next_index,[]) p
      ~f:(fun (index,next_index,acc) (vsrc,vdst) ->
          match Vertex.Map.find index vsrc with
          | Some (i,vdst') ->
            if compare_vertex vdst' vdst = 0
            then (index,next_index,i::acc)
            else invalid_arg
                (Format.asprintf "add_facets: %a has %a and %a as destination"
                   pp_vertex vsrc pp_vertex vdst pp_vertex vdst')
          | None -> (Vertex.Map.add index vsrc (next_index,vdst),next_index+1,next_index::acc)
        ) in
  {index;next_index;facets=(List.rev acc)::s.facets}

let get_solution s =
  let mapping = Vertex.Map.to_alist s.index in
  let mapping = List.sort mapping  ~cmp:(fun (_,((i:int),_)) (_,(j,_)) -> compare i j) in
  let mapping = List.map ~f:(fun (src,(_,dst)) -> src,dst ) mapping in
  {mapping;facets=s.facets}

let get_destination_facets sol =
  let mapping = Int.Table.create () in
  List.iteri sol.mapping ~f:(fun i map ->
      Int.Table.add_exn mapping i map);
  let get_dst id = snd (Int.Table.find_exn mapping id) in
  let sol = List.map sol.facets ~f:(fun l -> List.map ~f:(get_dst) l) in
  sol

let compute_resemblance shape sol =
  let sol = get_destination_facets sol in
  let side = Geom_ops.get_shap_sides_two shape sol in
  let union,inter =
    Geom_ops.Vertex.Table.fold ~init:(Q.zero,Q.zero)
      side ~f:(fun ~key:_ ~data:(v1,v2,v3,(c1,c2)) (union,inter) ->
        let a = Q.abs (Geom_ops.area [v1;v2;v3]) in
        let union = if c1 || c2 then Q.(union + a) else union in
        let inter = if c1 && c2 then Q.(inter + a) else inter in
        union,inter
      ) in
  let r = (Q.div inter union) in
  let approx = 1000000 in
  let r = Q.to_bigint (Q.(mul r (Q.of_int approx))) in
  let r = (Z.to_float r) /. (float approx) in
  r

let compute_resemblance shape sol =
  Or_error.try_with_join ~backtrace:true
    (fun () ->
       let r = compute_resemblance shape sol in
       if Float.(0. <= r && r <= 1.) then Ok r
       else Or_error.errorf "outside range: %f" r
    )

let renormalize_solution solution =
  let a = Array.of_list solution.mapping in
  let sol =
    List.fold_left solution.facets
      ~f:(fun sol facet ->
          let facet = List.map ~f:(fun i -> Array.get a i) facet in
           add_facets facet sol)
      ~init:empty_solution
  in
  get_solution sol
