open Core.Std;;
open Geom_ops;;
open Quompare;;

(* chargement d'un problème *)
let problem_directory = "../Problems";;
let problem_file id = Filename.concat problem_directory (Printf.sprintf "%04i.origami" id);;
let parse_problem i = Icfp_parse.parse (In_channel.read_all (problem_file i));;

(* Recherche d'un depliage *)
type step =
   { list_points : (vertex, vertex) Hashtbl.Poly.t;
     list_voisins : (vertex, vertex list) Hashtbl.Poly.t; }
   [@@deriving sexp]

let copy_step last_step : step = { list_points = Hashtbl.Poly.copy last_step.list_points ;
                                   list_voisins = Hashtbl.Poly.copy last_step.list_voisins };;

let init_unfold (segment_list : Geom_ops.segment list) : step =
  let points = Hashtbl.Poly.create () in
  let voisins  = Hashtbl.Poly.create () in
  let init p1 p2 =
    Hashtbl.Poly.add_multi voisins ~key:p1 ~data:p2;
    Hashtbl.Poly.set points ~key:p1 ~data:p1
  in
  List.iter segment_list (fun (S(p1,p2)) -> init p1 p2; init p2 p1);
  { list_points = points ; list_voisins = voisins };;

let find_left_point
    (last_pt:Geom_ops.vertex) (cur_pt:Geom_ops.vertex)
    (pt1:Geom_ops.vertex) (pt2:Geom_ops.vertex) =
  let zone produit_scalaire produit_vectoriel =
    match Q.(produit_scalaire > Q.of_int 0, produit_vectoriel > Q.of_int 0) with
    | (false, false) -> 3
    | (false, true) -> 0
    | (true, false) -> 2
    | (true, true) -> 1 in
  let vo = cur_pt -= last_pt in
  let v1 = pt1 -= cur_pt in
  let v2 = pt2 -= cur_pt in
  let scal1 = Geom_ops.scal vo v1 in
  let scal2 = Geom_ops.scal vo v2 in
  let vect1 = Geom_ops.vect_prod vo v1 in
  let vect2 = Geom_ops.vect_prod vo v2 in
  let zone1 = zone scal1 vect1 in
  let zone2 = zone scal2 vect2 in
  if Int.(zone1 < zone2) then pt1
  else if Int.(zone2 < zone1) then pt2
  else if Geom_ops.vect_prod v1 v2 > Q.of_int 0 then pt2
  else pt1

let facettes_from_pt (current_step:step) (pt:Geom_ops.vertex) =
  let rec loop
      (cur_pt:Geom_ops.vertex)
      (last_pt:Geom_ops.vertex) (acc:Geom_ops.vertex list) =
    if compare_vertices cur_pt pt = 0 then
      List.rev acc
    else
      let list_pt_futurs = Hashtbl.Poly.find_exn current_step.list_voisins cur_pt in
      let list_pt_futurs_reduit =
        List.filter list_pt_futurs (fun v -> compare_vertices v last_pt <> 0)
      in
      let next_pt = match list_pt_futurs_reduit with
        | [] -> assert false
        | h::t ->
          List.fold t ~init:h ~f:(find_left_point last_pt cur_pt)
      in
      loop next_pt cur_pt (next_pt::acc)
  in
  let list_voisins = Hashtbl.Poly.find_exn current_step.list_voisins pt in
  List.map list_voisins ~f:(fun (voisin : Geom_ops.vertex) -> loop voisin pt [voisin])

let mk_facettes_ccw (current_step:step) (pt:Geom_ops.vertex) =
  List.filter (facettes_from_pt current_step pt) ~f:ccw

let mk_facettes_cw (current_step:step) (pt:Geom_ops.vertex) =
  List.filter (facettes_from_pt current_step pt) ~f:(fun x -> not (ccw x))

let mk_cycles_cw (current_step:step) (pt:Geom_ops.vertex) =
  let rec loop chemins_fini chemins_en_cours = match chemins_en_cours with
    | [] ->
      chemins_fini
    | (h::t)::rest ->
      let list_voisins = Hashtbl.Poly.find_exn current_step.list_voisins h in
      let equal_vertex p1 p2 = compare_vertex p1 p2 = 0 in
      let list_futurs = List.filter list_voisins ~f:(fun voisin -> not (List.exists (pt::t) (equal_vertex voisin))) in
      let nouveaux_chemin = List.map list_futurs ~f:(fun futur -> futur::h::t) in
      let nouveaux_fini = if List.exists list_voisins (equal_vertex pt) then (h::t)::chemins_fini else chemins_fini in
      loop nouveaux_fini (List.rev_append nouveaux_chemin rest)
    | _ -> assert false in
  List.filter (loop [] [[pt]]) ~f:ccw

let pts_of_plis (current_step : step) (pli : Geom_ops.segment) : Geom_ops.vertex list =
  let pli = representant_segment pli in
  let S(origine,pt_courant) = pli in
  let ligne = L(pt_courant,pt_courant-=origine) in
  let list_voisins = Hashtbl.Poly.find_exn current_step.list_voisins pt_courant in
  let voisins_reduit = List.filter list_voisins (fun x -> compare_vertices x pt_courant <> 0) in
  List.filter voisins_reduit ~f:(fun cible ->
      let symetrique = symetrie ligne cible in
      let symetrique_vector = symetrique -= pt_courant in
      let test align = (vect_prod symetrique_vector (align-=pt_courant)) = Q.of_int 0 in
      List.exists list_voisins ~f:test)

let mk_all_plis (current_step : step) =
  let list_points = Hashtbl.Poly.keys current_step.list_points in
  let list2_plis = List.map list_points (fun origine ->
      let list_voisins = Hashtbl.Poly.find_exn current_step.list_voisins origine in
      List.map list_voisins (fun destination -> representant_segment (S(origine,destination)))) in
  let list_plis = Caml.List.flatten list2_plis in
  List.dedup ~compare:Geom_ops.compare_segment list_plis

let mk_border_plis (current_step : step) (border_point : Geom_ops.vertex) =
  match mk_facettes_cw current_step border_point with
  | x::[] -> let (first_point,second_point,rest) = match x with
      | p1::p2::r -> (p1,p2,r)
      | _ -> assert false in
    let rec to_segment_list prev rest acc = match rest with
      | [] -> acc
      | h::t -> to_segment_list h t ((S(prev,h))::acc) in
    to_segment_list first_point (second_point::rest) [(S(first_point,second_point))]
  | _ -> assert false
  
let uniquify facet =
  to_list (representant (to_dll facet))

let dedup_facets facets =
  List.dedup (List.map facets uniquify)

let plis_to_facettes (current_step : step) (datas_plis : ('a * Geom_ops.vertex list) Core.Std.List.t) =
  let memoize = Caml.Hashtbl.create 100 in
  let mk_facettes v =
    match Caml.Hashtbl.find memoize v with
    | exception Not_found ->
      let res = mk_facettes_ccw current_step v in
      Caml.Hashtbl.add memoize v res;
      res
    | res -> res in
  let resultat_with_duplicate = List.map datas_plis ~f:(fun (pli, list_points) ->
      let list_facette = List.join (List.map list_points ~f:mk_facettes) in
      pli,list_facette) in
  let resultat_without_duplicate = List.map resultat_with_duplicate ~f:(fun (pli, list_facette) -> pli, dedup_facets list_facette)  in
  Caml.List.flatten (List.map ~f:(fun (p,l) -> List.map ~f:(fun a -> p,a) l) resultat_without_duplicate)

let mk_list_next_steps (current_step : step) (filter_pts : bool) (filter_pli_border : bool) =
  let list_points = Hashtbl.Poly.keys current_step.list_points in
  let border_point = match list_points with
    | [] -> assert false
    | h::t -> List.fold t ~init:h ~f:(fun p1 p2 -> if Int.(compare_vertex p1 p2 < 0) then p1 else p2) in
  let all_cycles = mk_cycles_cw current_step border_point in
  let list_pli =
    if filter_pli_border
    then mk_border_plis current_step border_point
    else mk_all_plis current_step in
  if filter_pts then
    plis_to_facettes current_step @@ List.map list_pli ~f:(fun pli -> pli,pts_of_plis current_step pli)
  else
    List.join @@ List.map list_pli ~f:(fun pli -> List.map all_cycles ~f:(fun cycle -> pli,cycle))

let unfold (current_step : step) (add_step : step -> unit) ((pli,facette) : Geom_ops.segment * Geom_ops.vertex list) : 'a option
           = None

let proceed (profondeur_max : int) (filter_pts : bool) (filter_pli_border : bool) (id_problem : int) =
  let problem = parse_problem id_problem in
  let (_, squelette) = problem in
  let step0 = init_unfold squelette in
  let todo_stack = Stack.create () in
  let _ = Stack.push todo_stack (step0, profondeur_max) in
  let rec loop () =
    let current_step = Stack.pop todo_stack in
    match current_step with
    | None -> None
    | Some (current_step, profondeur_max) ->
      let next_steps = mk_list_next_steps current_step filter_pts filter_pli_border in
      let add_step new_step = Stack.push todo_stack (new_step, profondeur_max - 1) in
      let result = List.fold next_steps ~init:None ~f:(fun acc x -> match acc with None -> unfold current_step add_step x | _ -> acc) in
      match result with
      | None -> loop ()
      | _ -> result in
  let proceed = loop () in
  proceed;;

let result = proceed 3 true true 11;;



let deplie_facet facet ((S(p1, p2)) as pli) =
  let l = L(p1,p2 -= p1) in
  List.filter_map ~f:(fun ((S(f1,f2)) as seg) ->
      if equal_segment seg pli then
        None
      else
        let s = S(symetrie l f1, symetrie l f2) in
        Some s)
    facet
  
(* let mk_list_triangle (list_polygone : Geom_ops.vertex list list) = *)
(*   let rec reduce_dll (list_triangle : (Geom_ops.vertex * Geom_ops.vertex * Geom_ops.vertex) list) (dll_poly : Geom_ops.vertex Geom_ops.dll) (nb_sommet : int) = *)
(*     let new_triangle = (dll_poly.contents, dll_poly.next.contents, dll_poly.prev.contents) in *)
(*     if Int.(nb_sommet < 3) then assert false else *)
(*     if nb_sommet = 3 then new_triangle::list_triangle else *)
(*       let current = dll_poly.contents in *)
(*       let next = dll_poly.next.contents in *)
(*       let prev = dll_poly.prev.contents in *)
(*       if (Geom_ops.vect_prod (next-=current) (prev-=current)) > Q.of_int 0 then *)
(*         reduce_dll (new_triangle::list_triangle) (Geom_ops.shortcut dll_poly.prev dll_poly.next) (nb_sommet - 1) else *)
(*         reduce_dll list_triangle dll_poly.next nb_sommet in *)
(*   let reduce_polygone (list_triangle : (Geom_ops.vertex * Geom_ops.vertex * Geom_ops.vertex) list) (polygone : Geom_ops.vertex list) = *)
(*     let nb_sommet = List.length polygone in *)
(*     let dll_poly = Geom_ops.to_dll polygone in *)
(*     reduce_dll list_triangle dll_poly nb_sommet in *)
(*   let rec reduce_polynomes list_positive list_negative rest = match rest with *)
(*     | []   -> list_positive,list_negative *)
(*     | h::t -> let sens_h = Geom_ops.ccw h in *)
(*       let new_list_positive = if sens_h then reduce_polygone list_positive h else list_positive in *)
(*       let new_list_negative = if sens_h then list_negative else reduce_polygone list_negative (List.rev h) in *)
(*       reduce_polynomes new_list_positive new_list_negative t *)
(*   in reduce_polynomes [] [] list_polygone *)

(* let pt_is_in_surface pt (list_triangle_pos,list_triangle_neg) = *)
(*   let pt_is_in_triangle ((a,b,c) : Geom_ops.vertex * Geom_ops.vertex * Geom_ops.vertex) = *)
(*     let cote pta ptb = *)
(*       let vab = ptb -= pta in *)
(*       let vap = pt  -= pta in *)
(*       Geom_ops.vect_prod vab vap >= Q.of_int 0 in *)
(*     let coteA = cote a b in *)
(*     let coteB = cote b c in *)
(*     let coteC = cote c a in *)
(*     coteA && coteB && coteC in *)
(*   if List.exists list_triangle_neg pt_is_in_triangle then false *)
(*   else if List.exists list_triangle_pos pt_is_in_triangle then true *)
(*   else false *)

(* let polygones = [] *)
  (* [[p 0 0; p 0 1; p 1 0]] *)
let a, b, c, d =
  p 0 0, p 0 1, p 1 0, p 1 1

let e =
  { x = Q.make (Z.of_int 1) (Z.of_int 3);
    y = Q.of_int 1 }
(* let squelette = *)
(*   [S (a, b); *)
(*    S (b, c); *)
(*    S (c, a); *)
(*    S (b, d); *)
(*    S (d, c); *)
(*   ] *)

(* let problem : Icfp_parse.problem = polygones, squelette *)
(* let triangles = mk_list_triangle polygones;; *)

let () =
  Format.printf "%a@.@." Sexp.pp_hum (sexp_of_step step0)

type plisage = (Geom_ops.segment * Geom_ops.vertex list) list
  [@@deriving sexp]

let list_plis = mk_list_plis step0;;

let () =
  Format.printf "plis:@ %a@.@." Sexp.pp_hum (sexp_of_plisage list_plis)

type morf = Geom_ops.vertex Core.Std.List.t Core.Std.List.t
  [@@deriving sexp]
type chose = Geom_ops.vertex Core.Std.List.t
  [@@deriving sexp]

(* let exemple_pli = *)
(*   match Hashtbl.Poly.keys list_plis with [] -> assert false | h::_ -> h *)
(* let exemple_point = *)
(*   match Hashtbl.Poly.find_exn list_plis exemple_pli with [] -> assert false | h::_ -> h *)
(* let list_facette = mk_facettes step0 exemple_point;; *)

let list_voisins = Hashtbl.Poly.find_exn step0.list_voisins e

let () =
  Format.printf "l: %i@." (List.length all_facets)

let () =
  Format.printf "voisins:@ %a@.@." Sexp.pp_hum (sexp_of_chose list_voisins)

let ccw(* , cw *) = mk_facettes step0 e

let () =
  Format.printf "ccw:@ %a@.@." Sexp.pp_hum (sexp_of_morf ccw)

(* let () = *)
(*   Format.printf "cw:@ %a@.@." Sexp.pp_hum (sexp_of_morf cw) *)
